# COPYRIGHT (C) 2020-2024 Nicotine+ Contributors
# COPYRIGHT (C) 2008-2012 quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2007-2009 daelstorm <daelstorm@gmail.com>
# COPYRIGHT (C) 2003-2004 Hyriand <hyriand@thegraveyard.org>
# COPYRIGHT (C) 2001-2003 Alexander Kanavin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import random
import selectors
import socket
import struct
import sys
import time

from collections import defaultdict
from collections import deque
from os import strerror
from threading import Thread

from pynicotine.events import events
from pynicotine.logfacility import log
from pynicotine.slskmessages import DISTRIBUTED_MESSAGE_CLASSES
from pynicotine.slskmessages import DISTRIBUTED_MESSAGE_CODES
from pynicotine.slskmessages import NETWORK_MESSAGE_EVENTS
from pynicotine.slskmessages import PEER_MESSAGE_CLASSES
from pynicotine.slskmessages import PEER_MESSAGE_CODES
from pynicotine.slskmessages import PEER_INIT_MESSAGE_CLASSES
from pynicotine.slskmessages import PEER_INIT_MESSAGE_CODES
from pynicotine.slskmessages import SERVER_MESSAGE_CLASSES
from pynicotine.slskmessages import SERVER_MESSAGE_CODES
from pynicotine.slskmessages import DOUBLE_UINT32_UNPACK
from pynicotine.slskmessages import UINT32_UNPACK
from pynicotine.slskmessages import AcceptChildren
from pynicotine.slskmessages import BranchLevel
from pynicotine.slskmessages import BranchRoot
from pynicotine.slskmessages import CantConnectToPeer
from pynicotine.slskmessages import CloseConnection
from pynicotine.slskmessages import ConnectionType
from pynicotine.slskmessages import ConnectToPeer
from pynicotine.slskmessages import DistribBranchLevel
from pynicotine.slskmessages import DistribBranchRoot
from pynicotine.slskmessages import DistribEmbeddedMessage
from pynicotine.slskmessages import DistribSearch
from pynicotine.slskmessages import DownloadFile
from pynicotine.slskmessages import EmbeddedMessage
from pynicotine.slskmessages import EmitNetworkMessageEvents
from pynicotine.slskmessages import FileOffset
from pynicotine.slskmessages import FileSearchResponse
from pynicotine.slskmessages import FileTransferInit
from pynicotine.slskmessages import GetPeerAddress
from pynicotine.slskmessages import GetUserStats
from pynicotine.slskmessages import GetUserStatus
from pynicotine.slskmessages import HaveNoParent
from pynicotine.slskmessages import Login
from pynicotine.slskmessages import MessageType
from pynicotine.slskmessages import PossibleParents
from pynicotine.slskmessages import ParentMinSpeed
from pynicotine.slskmessages import ParentSpeedRatio
from pynicotine.slskmessages import PeerInit
from pynicotine.slskmessages import PierceFireWall
from pynicotine.slskmessages import Relogged
from pynicotine.slskmessages import ResetDistributed
from pynicotine.slskmessages import ServerConnect
from pynicotine.slskmessages import ServerDisconnect
from pynicotine.slskmessages import ServerReconnect
from pynicotine.slskmessages import SetDownloadLimit
from pynicotine.slskmessages import SetUploadLimit
from pynicotine.slskmessages import SetWaitPort
from pynicotine.slskmessages import SharedFileListResponse
from pynicotine.slskmessages import UnwatchUser
from pynicotine.slskmessages import UploadFile
from pynicotine.slskmessages import UserInfoResponse
from pynicotine.slskmessages import UserStatus
from pynicotine.slskmessages import WatchUser
from pynicotine.slskmessages import increment_token
from pynicotine.slskmessages import initial_token
from pynicotine.utils import human_speed


class Connection:
    __slots__ = ("sock", "addr", "io_events", "is_established", "in_buffer", "out_buffer",
                 "last_active", "recv_size")

    def __init__(self, sock=None, addr=None, io_events=None):

        self.sock = sock
        self.addr = addr
        self.io_events = io_events
        self.in_buffer = bytearray()
        self.out_buffer = bytearray()
        self.last_active = time.monotonic()
        self.recv_size = 51200
        self.is_established = False


class ServerConnection(Connection):
    __slots__ = ("login",)

    def __init__(self, *args, login=None, **kwargs):
        Connection.__init__(self, *args, **kwargs)
        self.login = login


class PeerConnection(Connection):
    __slots__ = ("init", "request_token", "response_token", "has_post_init_activity")

    def __init__(self, *args, init=None, request_token=None, response_token=None, **kwargs):

        Connection.__init__(self, *args, **kwargs)

        self.init = init
        self.request_token = request_token    # Requesting indirect connection to user
        self.response_token = response_token  # Responding to indirect connection request from user
        self.has_post_init_activity = False


class NetworkInterfaces:

    IP_BIND_ADDRESS_NO_PORT = SO_BINDTODEVICE = None

    if sys.platform == "win32":
        from ctypes import POINTER, Structure, wintypes

        AF_INET = 2

        GAA_FLAG_SKIP_ANYCAST = 2
        GAA_FLAG_SKIP_MULTICAST = 4
        GAA_FLAG_SKIP_DNS_SERVER = 8

        ERROR_BUFFER_OVERFLOW = 111

        class SockaddrIn(Structure):
            pass

        class SocketAddress(Structure):
            pass

        class IpAdapterUnicastAddress(Structure):
            pass

        class IpAdapterAddresses(Structure):
            pass

        SockaddrIn._fields_ = [  # pylint: disable=protected-access
            ("sin_family", wintypes.USHORT),
            ("sin_port", wintypes.USHORT),
            ("sin_addr", wintypes.BYTE * 4),
            ("sin_zero", wintypes.CHAR * 8)
        ]

        SocketAddress._fields_ = [  # pylint: disable=protected-access
            ("lp_sockaddr", POINTER(SockaddrIn)),
            ("i_sockaddr_length", wintypes.INT)
        ]

        IpAdapterUnicastAddress._fields_ = [  # pylint: disable=protected-access
            ("length", wintypes.ULONG),
            ("flags", wintypes.DWORD),
            ("next", POINTER(IpAdapterUnicastAddress)),
            ("address", SocketAddress)
        ]

        IpAdapterAddresses._fields_ = [  # pylint: disable=protected-access
            ("length", wintypes.ULONG),
            ("if_index", wintypes.DWORD),
            ("next", POINTER(IpAdapterAddresses)),
            ("adapter_name", wintypes.LPSTR),
            ("first_unicast_address", POINTER(IpAdapterUnicastAddress)),
            ("first_anycast_address", wintypes.LPVOID),
            ("first_multicast_address", wintypes.LPVOID),
            ("first_dns_server_address", wintypes.LPVOID),
            ("dns_suffix", wintypes.LPWSTR),
            ("description", wintypes.LPWSTR),
            ("friendly_name", wintypes.LPWSTR)
        ]

    elif sys.platform == "linux":
        IP_BIND_ADDRESS_NO_PORT = 24
        SIOCGIFADDR = 0x8915
        SO_BINDTODEVICE = 25

    elif sys.platform.startswith("sunos"):
        SIOCGIFADDR = -0x3fdf96f3  # Solaris

    else:
        SIOCGIFADDR = 0xc0206921   # macOS, *BSD

    @classmethod
    def _get_interface_addresses_win32(cls):
        """Returns a dictionary of network interface names and IP addresses (Win32).

        https://learn.microsoft.com/en-us/windows/win32/api/iphlpapi/nf-iphlpapi-getadaptersaddresses
        """

        # pylint: disable=invalid-name

        from ctypes import POINTER, byref, cast, create_string_buffer, windll, wintypes

        interface_addresses = {}
        adapter_addresses_size = wintypes.ULONG()
        return_value = cls.ERROR_BUFFER_OVERFLOW

        while return_value == cls.ERROR_BUFFER_OVERFLOW:
            p_adapter_addresses = cast(
                create_string_buffer(adapter_addresses_size.value), POINTER(cls.IpAdapterAddresses)
            )
            return_value = windll.Iphlpapi.GetAdaptersAddresses(
                cls.AF_INET,
                (cls.GAA_FLAG_SKIP_ANYCAST | cls.GAA_FLAG_SKIP_MULTICAST | cls.GAA_FLAG_SKIP_DNS_SERVER),
                None,
                p_adapter_addresses,
                byref(adapter_addresses_size),
            )

        if return_value:
            log.add_debug("Failed to get list of network interfaces. Error code %s", return_value)
            return interface_addresses

        while p_adapter_addresses:
            adapter_addresses = p_adapter_addresses.contents

            if adapter_addresses.first_unicast_address:
                interface_name = adapter_addresses.friendly_name
                socket_address = adapter_addresses.first_unicast_address[0].address
                interface_addresses[interface_name] = socket.inet_ntoa(socket_address.lp_sockaddr[0].sin_addr)

            p_adapter_addresses = adapter_addresses.next

        return interface_addresses

    @classmethod
    def _get_interface_addresses_posix(cls):
        """Returns a dictionary of network interface names and IP addresses
        (POSIX)"""

        interface_addresses = {}

        try:
            interface_name_index = socket.if_nameindex()

        except (AttributeError, OSError) as error:
            log.add_debug("Failed to get list of network interfaces: %s", error)
            return interface_addresses

        for _i, interface_name in interface_name_index:
            try:
                import fcntl

                with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
                    ip_interface = fcntl.ioctl(sock.fileno(),
                                               cls.SIOCGIFADDR,
                                               struct.pack("256s", interface_name.encode()[:15]))

                    ip_address = socket.inet_ntoa(ip_interface[20:24])
                    interface_addresses[interface_name] = ip_address

            except (ImportError, OSError) as error:
                log.add_debug("Failed to get IP address for network interface %s: %s", (interface_name, error))
                continue

        return interface_addresses

    @classmethod
    def get_interface_addresses(cls):
        """Returns a dictionary of network interface names and IP addresses."""

        if sys.platform == "win32":
            return cls._get_interface_addresses_win32()

        return cls._get_interface_addresses_posix()

    @classmethod
    def get_interface_address(cls, interface_name):
        """Returns the IP address of a specific network interface."""

        if not interface_name:
            return None

        return cls.get_interface_addresses().get(interface_name)

    @classmethod
    def bind_to_interface(cls, sock, interface_name, address):
        """Bind socket to the specified network interface name, if required on
        the current platform. Otherwise bind to the IP address of the network
        interface, retrieved from get_interface_addresses().
        """

        if cls.SO_BINDTODEVICE:
            try:
                # We need to use SO_BINDTODEVICE on Linux, since socket.bind() has no
                # effect on routing (weak host model).
                sock.setsockopt(socket.SOL_SOCKET, cls.SO_BINDTODEVICE, interface_name.encode())
                return

            except PermissionError as error:
                log.add_debug("Failed to bind socket to network interface with SO_BINDTODEVICE. "
                              "Falling back to socket.bind(). Error: %s", error)

                # We only need to bind to the interface address, not a port.
                # Set IP_BIND_ADDRESS_NO_PORT to avoid reserving an ephemeral port in bind(),
                # and let connect() select one instead later.
                sock.setsockopt(socket.SOL_IP, cls.IP_BIND_ADDRESS_NO_PORT, 1)

        sock.bind((address, 0))


class NetworkThread(Thread):
    """This is the networking thread that does all the communication with the
    Soulseek server and peers. Communication with the core is done through
    events.

    The server and peers send each other small binary messages that
    start with length and message code followed by the actual message
    data.
    """

    __slots__ = ("pending_shutdown", "upload_speed", "token", "_pending_network_msgs",
                 "_user_update_counter", "_user_update_counters", "_upload_queue_timer_id",
                 "_retry_failed_uploads_timer_id")

    IN_PROGRESS_STALE_AFTER = 2
    INDIRECT_REQUEST_TIMEOUT = 20
    CONNECTION_MAX_IDLE = 60
    CONNECTION_MAX_IDLE_GHOST = 10
    CONNECTION_BACKLOG_LENGTH = 65535      # OS limit can be lower
    MAX_INCOMING_MESSAGE_SIZE = 469762048  # 448 MiB, to leave headroom for large shares
    ALLOWED_PEER_CONN_TYPES = {
        ConnectionType.PEER,
        ConnectionType.FILE,
        ConnectionType.DISTRIBUTED
    }
    ERROR_NOT_CONNECTED = OSError(errno.ENOTCONN, strerror(errno.ENOTCONN))
    ERROR_TIMED_OUT = OSError(errno.ETIMEDOUT, strerror(errno.ETIMEDOUT))

    # Looping max ~240 times per second (SLEEP_MIN_IDLE) on high activity
    # ~20 (SLEEP_MAX_IDLE + SLEEP_MIN_IDLE) by default
    SLEEP_MAX_IDLE = 0.04584
    SLEEP_MIN_IDLE = 0.00416

    try:
        import resource

        # Increase the process file limit to a maximum of 10240 (macOS limit), to provide
        # breathing room for opening both peer sockets and regular files (file transfers,
        # log files etc.)

        _SOFT_FILE_LIMIT, HARD_FILE_LIMIT = resource.getrlimit(resource.RLIMIT_NOFILE)    # pylint: disable=no-member
        MAX_FILE_LIMIT = min(HARD_FILE_LIMIT, 10240)

        resource.setrlimit(resource.RLIMIT_NOFILE, (MAX_FILE_LIMIT, MAX_FILE_LIMIT))  # pylint: disable=no-member

        # Reserve 2/3 of the file limit for sockets, but always limit the maximum number
        # of sockets to 3072 to improve performance.

        MAX_SOCKETS = min(int(MAX_FILE_LIMIT * (2 / 3)), 3072)

    except ImportError:
        # For Windows, FD_SETSIZE is set to 512 in CPython.
        # This limit is hardcoded, so we'll have to live with it for now.
        # https://github.com/python/cpython/issues/72894

        MAX_SOCKETS = 512

    def __init__(self):

        super().__init__(name="NetworkThread")

        self._message_queue = deque()
        self._pending_peer_conns = {}
        self._pending_init_msgs = defaultdict(list)
        self._token_init_msgs = {}
        self._username_init_msgs = {}
        self._user_addresses = {}
        self._should_process_queue = False
        self._want_abort = False

        self._selector = None
        self._listen_socket = None
        self._listen_port = None
        self._interface_name = None
        self._interface_address = None
        self._portmapper = None
        self._local_ip_address = ""

        self._server_conn = None
        self._server_address = None
        self._server_username = None
        self._server_timeout_time = None
        self._server_timeout_value = -1
        self._manual_server_disconnect = False
        self._manual_server_reconnect = False
        self._server_relogged = False

        self._parent_conn = None
        self._potential_parents = {}
        self._child_peers = {}
        self._branch_level = 0
        self._branch_root = None
        self._is_server_parent = False
        self._distrib_parent_min_speed = 0
        self._distrib_parent_speed_ratio = 1
        self._max_distrib_children = 0
        self._upload_speed = 0

        self._num_sockets = 0
        self._last_cycle_time = 0

        self._conns = {}
        self._token = initial_token()

        self._file_init_msgs = {}
        self._file_download_msgs = {}
        self._file_upload_msgs = {}
        self._conns_downloaded = defaultdict(int)
        self._conns_uploaded = defaultdict(int)
        self._calc_upload_limit_function = self._calc_upload_limit_none
        self._upload_limit = 0
        self._download_limit = 0
        self._upload_limit_split = 0
        self._download_limit_split = 0
        self._total_uploads = 0
        self._total_downloads = 0
        self._total_download_bandwidth = 0
        self._total_upload_bandwidth = 0

        for event_name, callback in (
            ("enable-message-queue", self._enable_message_queue),
            ("queue-network-message", self._queue_network_message),
            ("schedule-quit", self._schedule_quit),
            ("start", self.start)
        ):
            events.connect(event_name, callback)

    def _enable_message_queue(self):
        self._should_process_queue = True

    def _queue_network_message(self, msg):
        if self._should_process_queue:
            self._message_queue.append(msg)

    def _schedule_quit(self):
        self._want_abort = True

    # Listening Socket #

    def _create_listen_socket(self):

        self._listen_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._listen_socket.setblocking(False)
        self._num_sockets += 1

        # On platforms other than Windows, SO_REUSEADDR is necessary to allow binding
        # to the same port immediately after reconnecting. This option behaves differently
        # on Windows, allowing other programs to hijack the port, so don't set it there.

        if sys.platform != "win32":
            self._listen_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        if not self._bind_listen_port():
            self._close_listen_socket()
            return False

        self._selector.register(self._listen_socket, selectors.EVENT_READ)
        return True

    def _close_listen_socket(self):

        if self._listen_socket is None:
            return

        try:
            self._selector.unregister(self._listen_socket)

        except KeyError:
            # Socket was not registered
            pass

        self._close_socket(self._listen_socket)
        self._listen_socket = None
        self._listen_port = None
        self._num_sockets -= 1

    def _bind_listen_port(self):

        if not self._bind_socket_interface(self._listen_socket):
            self._set_server_timer(use_fixed_timeout=True)
            log.add(_("Specified network interface '%s' is not available"), self._interface_name)
            return False

        try:
            ip_address = self._interface_address or self._find_local_ip_address()

            self._listen_socket.bind((ip_address, self._listen_port))
            self._listen_socket.listen(self.CONNECTION_BACKLOG_LENGTH)

        except OSError as error:
            self._set_server_timer(use_fixed_timeout=True)
            log.add(_("Cannot listen on port %(port)s. Ensure no other application uses it, or choose a "
                      "different port. Error: %(error)s"), {"port": self._listen_port, "error": error})
            self._listen_port = None
            return False

        self._local_ip_address = ip_address

        if self._interface_name:
            log.add_debug("Network interface: %s", self._interface_name)

        log.add_debug("Local IP address: %s", ip_address)
        log.add_debug("Maximum number of concurrent connections (sockets): %s", self.MAX_SOCKETS)
        log.add(_("Listening on port: %i"), self._listen_port)
        return True

    # Connections #

    def _indirect_request_error(self, token, init):

        username = init.target_user
        conn_type = init.conn_type

        log.add_conn("Indirect connect request of type %s to user %s with "
                     "token %s failed", (conn_type, username, token))

        if init.sock is not None:
            return

        # No direct connection was established, give up
        events.emit_main_thread(
            "peer-connection-error", username=username, conn_type=conn_type,
            msgs=init.outgoing_msgs[:]
        )
        init.outgoing_msgs.clear()
        self._username_init_msgs.pop(username + conn_type, None)

    def _check_indirect_request_timeouts(self, current_time=None, expire_all=False):

        if not self._token_init_msgs:
            return

        timed_out_requests = set()

        for token, (init, request_time) in self._token_init_msgs.items():
            if not expire_all and (current_time - request_time) < self.INDIRECT_REQUEST_TIMEOUT:
                continue

            self._indirect_request_error(token, init)
            timed_out_requests.add(token)

        if not timed_out_requests:
            return

        for token in timed_out_requests:
            del self._token_init_msgs[token]

        timed_out_requests.clear()

    def _is_connection_still_active(self, conn):

        init = conn.init

        if init is not None and (init.conn_type != "P" or init.target_user == self._server_username):
            # Distributed and file connections, as well as connections to ourselves,
            # are critical. Always assume they are active.
            return True

        return len(conn.out_buffer) > 0 or len(conn.in_buffer) > 0

    def _bind_socket_interface(self, sock):
        """Attempt to bind socket to an IP address, if provided with the
        --bindip CLI argument. Otherwise retrieve the IP address of the
        requested interface name, cache it for later, and bind to it.
        """

        if self._interface_address:
            if sock is not self._listen_socket:
                NetworkInterfaces.bind_to_interface(sock, self._interface_name, self._interface_address)

            return True

        if not self._interface_name:
            return True

        return False

    def _find_local_ip_address(self):

        # Create a UDP socket
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as local_socket:

            # Send a broadcast packet on a local address (doesn't need to be reachable,
            # but MacOS requires port to be non-zero)
            local_socket.connect_ex(("10.255.255.255", 1))

            # This returns the "primary" IP on the local box, even if that IP is a NAT/private/internal IP
            ip_address = local_socket.getsockname()[0]

        return ip_address

    def _add_init_message(self, init):

        conn_type = init.conn_type

        if conn_type == ConnectionType.FILE:
            # File transfer connections are not unique or reused later
            return True

        init_key = init.target_user + conn_type

        if init_key not in self._username_init_msgs:
            self._username_init_msgs[init_key] = init
            return True

        return False

    @staticmethod
    def _pack_network_message(msg):

        try:
            return msg.make_network_message()

        except Exception:
            from traceback import format_exc
            log.add("Unable to pack message type %s: %s", (msg.__class__, format_exc()))

        return None

    @staticmethod
    def _unpack_network_message(msg_class, msg_content, msg_size, conn_type, sock=None, addr=None, username=None):

        try:
            msg = msg_class()

            if sock is not None:
                msg.sock = sock

            if addr is not None:
                msg.addr = addr

            if username is not None:
                msg.username = username

            msg.parse_network_message(msg_content)
            return msg

        except Exception as error:
            log.add_debug("Unable to parse %s message type %s, size %s, contents %s. Error: %s",
                          (conn_type, msg_class, msg_size, msg_content, error))

        return None

    @staticmethod
    def _unpack_embedded_message(msg):
        """This message embeds a distributed message.

        We unpack the distributed message and process it.
        """

        msg_type = msg.distrib_code

        if msg_type not in DISTRIBUTED_MESSAGE_CLASSES:
            log.add_debug("Embedded distrib message type %s unknown", msg_type)
            return None

        distrib_class = DISTRIBUTED_MESSAGE_CLASSES[msg_type]
        unpacked_msg = distrib_class()
        unpacked_msg.parse_network_message(memoryview(msg.distrib_message))

        return unpacked_msg

    def _emit_network_message_event(self, msg):

        if msg is None:
            return

        msg_class = msg.__class__
        log.add_msg_contents(msg)

        if msg_class in NETWORK_MESSAGE_EVENTS:
            event_name = NETWORK_MESSAGE_EVENTS[msg_class]
            events.emit_main_thread(event_name, msg)

    def _modify_connection_events(self, conn, io_events):

        if conn.io_events != io_events:
            self._selector.modify(conn.sock, io_events)
            conn.io_events = io_events

    def _process_conn_messages(self, init):
        """A connection is established with the peer, time to queue up our peer
        messages for delivery."""

        username = init.target_user
        sock = init.sock
        msgs = init.outgoing_msgs

        for j in msgs:
            j.username = username
            j.sock = sock

        self._process_outgoing_messages(msgs)
        msgs.clear()

    def _send_message_to_peer(self, username, msg):

        conn_type = msg.msg_type

        if conn_type not in self.ALLOWED_PEER_CONN_TYPES:
            log.add_conn("Unknown connection type %s", conn_type)
            return

        init = None
        init_key = username + conn_type

        # Check if there's already a connection for the specified username
        if init_key in self._username_init_msgs:
            init = self._username_init_msgs[init_key]

        if init is None and conn_type != ConnectionType.FILE and username in self._pending_init_msgs:
            # Check if we have a pending PeerInit message (currently requesting user IP address)
            for pending_init in self._pending_init_msgs[username]:
                if pending_init.conn_type == conn_type:
                    init = pending_init
                    break

        if init is not None:
            log.add_conn("Sending message of type %s to user %s on existing connection",
                         (msg.__class__, username))

            init.outgoing_msgs.append(msg)

            if init.sock is not None and self._conns[init.sock].is_established:
                # We have initiated a connection previously, and it's ready
                self._process_conn_messages(init)

        else:
            log.add_conn("Sending message of type %s to user %s on new connection",
                         (msg.__class__, username))

            # This is a new peer, initiate a connection
            self._initiate_connection_to_peer(username, conn_type, msg)

    def _initiate_connection_to_peer(self, username, conn_type, msg=None, in_address=None):
        """Prepare to initiate a connection with a peer."""

        init = PeerInit(init_user=self._server_username, target_user=username, conn_type=conn_type)
        user_address = self._user_addresses.get(username)

        if in_address is not None:
            user_address = in_address

        elif user_address is not None:
            _ip_address, port = user_address

            if not port:
                # Port 0 means the user is likely bugged, ask the server for a new address
                user_address = None

        if msg is not None:
            init.outgoing_msgs.append(msg)

        if user_address is None:
            self._pending_init_msgs[username].append(init)
            self._send_message_to_server(GetPeerAddress(username))

            log.add_conn("Requesting address for user %s", username)
        else:
            self._connect_to_peer(username, user_address, init)

    def _connect_to_peer(self, username, addr, init, response_token=None):
        """Initiate a connection with a peer."""

        conn_type = init.conn_type

        if conn_type not in self.ALLOWED_PEER_CONN_TYPES:
            log.add_conn("Unknown connection type %s", conn_type)
            return

        if not self._add_init_message(init):
            log.add_conn("Direct connection of type %s to user %s (%s) requested, "
                         "but existing connection already exists", (conn_type, username, addr))
            return

        log.add_conn("Attempting direct connection of type %s to user %s, address %s",
                     (conn_type, username, addr))
        self._init_peer_connection(addr, init, response_token=response_token)

    def _connect_error(self, error, conn):

        if conn.__class__ is ServerConnection:
            server_address, port = conn.addr

            log.add(
                _("Cannot connect to server %(host)s:%(port)s: %(error)s"), {
                    "host": server_address,
                    "port": port,
                    "error": error
                }
            )
            self._set_server_timer()
            return

        conn_type = conn.init.conn_type
        username = conn.init.target_user
        response_token = conn.response_token

        if response_token is not None:
            log.add_conn("Cannot respond to indirect connection request of type %s from user %s, "
                         "token %s: %s", (conn_type, username, response_token, error))
            self._send_message_to_server(CantConnectToPeer(response_token, username))
            return

        log.add_conn("Direct connection of type %s to user %s failed: %s",
                     (conn_type, username, error))

    def _connect_to_peer_indirect(self, init):
        """Send a message to the server to ask the peer to connect to us
        (indirect connection)"""

        username = init.target_user
        conn_type = init.conn_type
        token = self._token = increment_token(self._token)
        request_time = time.monotonic()

        self._token_init_msgs[token] = (init, request_time)
        self._send_message_to_server(ConnectToPeer(token, username, conn_type))

        log.add_conn("Requesting indirect connection to user %s with token %s", (username, token))
        return token

    def _establish_outgoing_peer_connection(self, conn):

        conn.is_established = True
        init = conn.init
        sock = init.sock = conn.sock
        response_token = conn.response_token
        username = init.target_user
        conn_type = init.conn_type

        log.add_conn("Established outgoing connection of type %s with user %s. List of "
                     "outgoing messages: %s", (conn_type, username, init.outgoing_msgs))

        if response_token is not None:
            log.add_conn("Responding to indirect connection request of type %s from "
                         "user %s, token %s", (conn_type, username, response_token))
            self._process_outgoing_messages([PierceFireWall(sock, response_token)])
            self._accept_child_peer_connection(conn)
        else:
            log.add_conn("Sending peer init message of type %s to user %s", (conn_type, username))
            self._process_outgoing_messages([init])

        self._process_conn_messages(init)

    def _replace_existing_connection(self, init):

        username = init.target_user
        conn_type = init.conn_type

        if username == self._server_username:
            return

        prev_init = self._username_init_msgs.pop(username + conn_type, None)

        if prev_init is None or prev_init.sock is None:
            return

        log.add_conn("Discarding existing connection of type %s to user %s", (init.conn_type, username))

        init.outgoing_msgs = prev_init.outgoing_msgs
        prev_init.outgoing_msgs = []

        self._close_connection(self._conns[prev_init.sock])

    @staticmethod
    def _close_socket(sock):

        try:
            log.add_conn("Shutting down socket %s", sock)
            sock.shutdown(socket.SHUT_RDWR)

        except OSError as error:
            # Can't call shutdown if connection wasn't established, ignore error
            if error.errno != errno.ENOTCONN:
                log.add_conn("Failed to shut down socket %s: %s", (sock, error))

        log.add_conn("Closing socket %s", sock)
        sock.close()

    def _close_connection(self, conn):

        if conn is None:
            return

        sock = conn.sock
        del self._conns[sock]

        if conn is self._server_conn:
            # Disconnecting from server, clean up connections and queue
            self._server_disconnect()

        self._selector.unregister(sock)
        self._close_socket(sock)
        self._num_sockets -= 1

        conn.sock = None
        conn.in_buffer.clear()
        conn.out_buffer.clear()

        if conn.__class__ is not PeerConnection:
            return

        init = conn.init

        if init is None:
            # No peer init message present, nothing to do
            return

        conn_type = init.conn_type
        username = init.target_user
        addr = conn.addr
        is_connection_replaced = (init.sock is not sock)

        log.add_conn("Removed connection of type %s to user %s, address %s", (conn_type, username, addr))

        if not is_connection_replaced:
            init.sock = None

        if conn_type == ConnectionType.DISTRIBUTED:
            child_conn = self._child_peers.get(username)

            if child_conn is conn:
                self._remove_child_peer_connection(username)

            elif conn is self._parent_conn:
                self._send_have_no_parent()

        elif conn in self._file_init_msgs:
            file_init = self._file_init_msgs.pop(conn)

            if self._should_process_queue:
                timed_out = (time.monotonic() - conn.last_active) > self.CONNECTION_MAX_IDLE
                events.emit_main_thread(
                    "file-connection-closed", username=username, token=file_init.token,
                    sock=sock, timed_out=timed_out
                )

        if conn in self._file_download_msgs:
            del self._file_download_msgs[conn]
            self._total_downloads -= 1

            if not self._total_downloads:
                self._total_download_bandwidth = 0

            self._calc_download_limit()

        elif conn in self._file_upload_msgs:
            del self._file_upload_msgs[conn]
            self._total_uploads -= 1

            if not self._total_uploads:
                self._total_upload_bandwidth = 0

            self._calc_upload_limit_function()

        init_key = username + conn_type

        if init_key not in self._username_init_msgs:
            return

        log.add_conn("Removing peer init message of type %s for user %s, address %s",
                     (conn_type, username, addr))

        if is_connection_replaced or init is not self._username_init_msgs[init_key]:
            # Don't remove init message if connection has been superseded
            log.add_conn("Cannot remove peer init message, since the connection has been superseded")
            return

        if conn.request_token in self._token_init_msgs:
            # Indirect connection attempt in progress, remove init message later on timeout
            log.add_conn("Cannot remove peer init message, since an indirect connection attempt "
                         "is still in progress")
            return

        event_name = "peer-connection-closed" if conn.is_established else "peer-connection-error"
        events.emit_main_thread(
            event_name, username=username, conn_type=conn_type, msgs=init.outgoing_msgs[:])

        del self._username_init_msgs[init_key]

    def _is_connection_inactive(self, conn, current_time, num_sockets):

        if conn is self._server_conn:
            return False

        if num_sockets >= self.MAX_SOCKETS and not self._is_connection_still_active(conn):
            # Connection limit reached, close connection if inactive
            return True

        time_diff = (current_time - conn.last_active)

        if not conn.has_post_init_activity and time_diff > self.CONNECTION_MAX_IDLE_GHOST:
            # "Ghost" connections can appear when an indirect connection is established,
            # search results arrive, we close the connection, and the direct connection attempt
            # succeeds afterwrds. Since the peer already sent a search result message, this connection
            # idles without any messages ever being sent beyond PeerInit. Close it sooner than regular
            # idling connections to prevent connections from piling up.
            return True

        if time_diff > self.CONNECTION_MAX_IDLE:
            # No recent activity, peer connection is stale
            return True

        return False

    def _check_connections(self, current_time):

        num_sockets = self._num_sockets
        inactive_conns = set()
        stale_conns = set()

        for conn in self._conns.values():
            if not conn.is_established:
                if (current_time - conn.last_active) > self.IN_PROGRESS_STALE_AFTER:
                    stale_conns.add(conn)

            elif self._is_connection_inactive(conn, current_time, num_sockets):
                inactive_conns.add(conn)

            elif conn in self._file_download_msgs:
                file_download = self._file_download_msgs[conn]

                events.emit_main_thread(
                    "file-download-progress",
                    username=conn.init.target_user, token=file_download.token,
                    bytes_left=file_download.leftbytes, speed=file_download.speed
                )
                file_download.speed = 0

            elif conn in self._file_upload_msgs:
                file_upload = self._file_upload_msgs[conn]

                events.emit_main_thread(
                    "file-upload-progress",
                    username=conn.init.target_user, token=file_upload.token,
                    offset=file_upload.offset, bytes_sent=file_upload.sentbytes,
                    speed=file_upload.speed
                )
                file_upload.speed = 0

        if inactive_conns:
            for conn in inactive_conns:
                self._close_connection(conn)

            inactive_conns.clear()

        if stale_conns:
            for conn in stale_conns:
                self._connect_error(self.ERROR_TIMED_OUT, conn)
                self._close_connection(conn)

            stale_conns.clear()

        if self._pending_peer_conns:
            for addr, init in self._pending_peer_conns.copy().items():
                self._init_peer_connection(addr, init)

    # Server Connection #

    def _set_server_timer(self, use_fixed_timeout=False):

        if use_fixed_timeout:
            self._server_timeout_value = 5

        elif self._server_timeout_value == -1:
            # Add jitter to spread out connection attempts from Nicotine+ clients
            # in case server goes down
            self._server_timeout_value = random.randint(5, 15)

        elif 0 < self._server_timeout_value < 300:
            # Exponential backoff, max 5 minute wait
            self._server_timeout_value *= 2

        self._server_timeout_time = time.monotonic() + self._server_timeout_value
        log.add(_("Reconnecting to server in %s seconds"), self._server_timeout_value)

    @staticmethod
    def _set_server_socket_keepalive(sock, idle=10, interval=2):
        """Ensure we are disconnected from the server in case of connectivity
        issues, by sending TCP keepalive pings.

        Assuming default values are used, once we reach 10 seconds of
        idle time, we start sending keepalive pings once every 2
        seconds. If 10 failed pings have been sent in a row (20
        seconds), the connection is presumed dead.
        """

        count = 10
        timeout_seconds = (idle + (interval * count))

        if hasattr(socket, "SO_KEEPALIVE"):
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)  # pylint: disable=no-member

        if hasattr(socket, "TCP_KEEPINTVL"):
            sock.setsockopt(socket.IPPROTO_TCP,
                            socket.TCP_KEEPINTVL, interval)  # pylint: disable=no-member

        if hasattr(socket, "TCP_KEEPCNT"):
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, count)  # pylint: disable=no-member

        if hasattr(socket, "TCP_KEEPIDLE"):
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, idle)  # pylint: disable=no-member

        elif hasattr(socket, "TCP_KEEPALIVE"):
            # macOS fallback

            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPALIVE, idle)  # pylint: disable=no-member

        elif hasattr(socket, "SIO_KEEPALIVE_VALS"):
            # Windows fallback
            # Probe count is set to 10 on a system level, and can't be modified.
            # https://docs.microsoft.com/en-us/windows/win32/winsock/so-keepalive

            sock.ioctl(
                socket.SIO_KEEPALIVE_VALS,  # pylint: disable=no-member
                (
                    1,
                    idle * 1000,
                    interval * 1000
                )
            )

        if hasattr(socket, "TCP_USER_TIMEOUT"):
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_USER_TIMEOUT, timeout_seconds * 1000)

    def _server_connect(self, msg):
        """We're connecting to the server."""

        if self._server_conn is not None:
            return

        self._interface_name = msg.interface_name
        self._interface_address = (
            msg.interface_address or NetworkInterfaces.get_interface_address(self._interface_name)
        )
        self._listen_port = msg.listen_port

        if not self._create_listen_socket():
            self._should_process_queue = False
            events.emit_main_thread("set-connection-stats")  # Reset connection stats
            return

        self._portmapper = msg.portmapper

        self._manual_server_disconnect = False
        self._manual_server_reconnect = False
        self._server_timeout_time = None

        ip_address, port = msg.addr
        log.add(_("Connecting to %(host)s:%(port)s"), {"host": ip_address, "port": port})

        self._init_server_conn(msg)

    def _init_server_conn(self, msg):

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        io_events = selectors.EVENT_READ | selectors.EVENT_WRITE
        conn = ServerConnection(
            sock=sock, addr=msg.addr, io_events=io_events, login=msg.login
        )

        sock.setblocking(False)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        # Detect if our connection to the server is still alive
        self._set_server_socket_keepalive(sock)

        try:
            self._bind_socket_interface(sock)
            sock.connect_ex(msg.addr)

        except OSError as error:
            self._connect_error(error, conn)
            self._close_socket(sock)
            self._server_disconnect()
            return

        self._server_conn = self._conns[sock] = conn
        self._selector.register(sock, io_events)
        self._num_sockets += 1

    def _establish_outgoing_server_connection(self, conn):

        conn.is_established = True
        server_hostname, server_port = conn.addr

        log.add(
            _("Connected to server %(host)s:%(port)s, logging in…"), {
                "host": server_hostname,
                "port": server_port
            }
        )

        login, password = conn.login
        self._user_addresses[login] = (self._local_ip_address, self._listen_port)
        conn.login = True

        self._server_address = conn.addr
        self._server_username = self._branch_root = login
        self._server_timeout_value = -1

        self._send_message_to_server(
            Login(
                login, password,
                # Soulseek client version
                # NS and SoulseekQt use 157
                # We use a custom version number for Nicotine+
                160,

                # Soulseek client minor version
                # 17 stands for 157 ns 13c, 19 for 157 ns 13e
                # SoulseekQt seems to go higher than this
                # We use a custom minor version for Nicotine+
                2
            )
        )

        self._send_message_to_server(SetWaitPort(self._listen_port))

    def _process_server_message(self, msg_type, msg_size, in_buffer, start_offset, end_offset):

        msg_class = SERVER_MESSAGE_CLASSES[msg_type]
        msg = self._unpack_network_message(
            msg_class,
            memoryview(in_buffer)[start_offset:end_offset],
            msg_size,
            conn_type="server"
        )

        if msg is None:
            # Ignore unknown message and keep connection open
            return True

        if msg_class is EmbeddedMessage:
            self._distribute_embedded_message(msg)
            msg = self._unpack_embedded_message(msg)

        elif msg_class is Login:
            if msg.success:
                # Ensure listening port is open
                msg.local_address = self._user_addresses[self._server_username]
                local_ip_address, port = msg.local_address
                self._portmapper.set_port(port, local_ip_address)
                self._portmapper.add_port_mapping(blocking=True)

                msg.username = self._server_username
                msg.server_address = self._server_address

                # Ask for a list of parents to connect to (distributed network)
                self._send_have_no_parent()
            else:
                # Emit event and close connection
                self._emit_network_message_event(msg)
                return False

        elif msg_class is ConnectToPeer:
            username = msg.user
            addr = (msg.ip_address, msg.port)
            conn_type = msg.conn_type
            token = msg.token
            init = PeerInit(init_user=username, target_user=username, conn_type=conn_type)

            log.add_conn("Received indirect connection request of type %s from user %s, "
                         "token %s, address %s", (conn_type, username, token, addr))

            self._connect_to_peer(username, addr, init, response_token=token)

        elif msg_class is CantConnectToPeer:
            token = msg.token

            if token in self._token_init_msgs:
                init, _request_time = self._token_init_msgs.pop(token)
                self._indirect_request_error(token, init)

        elif msg_class is GetUserStatus:
            if msg.status == UserStatus.OFFLINE and msg.user in self._user_addresses:
                # User went offline, reset stored IP address
                self._user_addresses[msg.user] = None

        elif msg_class is GetPeerAddress:
            username = msg.user
            pending_init_msgs = self._pending_init_msgs.pop(msg.user, [])

            if not msg.port:
                log.add_conn("Server reported port 0 for user %s", username)

            addr = (msg.ip_address, msg.port)
            user_offline = (msg.ip_address == "0.0.0.0")

            for init in pending_init_msgs:
                # We now have the IP address for a user we previously didn't know,
                # attempt a connection with the peer/user
                if user_offline:
                    events.emit_main_thread(
                        "peer-connection-error", username=username, conn_type=init.conn_type,
                        msgs=init.outgoing_msgs[:], is_offline=True)
                else:
                    self._connect_to_peer(username, addr, init)

            # We already store a local IP address for our username
            if username != self._server_username and username in self._user_addresses:
                if user_offline or not msg.port:
                    addr = None

                self._user_addresses[username] = addr

        elif msg_class in (WatchUser, GetUserStats):
            if msg.user == self._server_username:
                if msg.avgspeed is not None:
                    self._upload_speed = msg.avgspeed
                    log.add_conn("Server reported our upload speed as %s", human_speed(msg.avgspeed))
                    self._update_maximum_distributed_children()

            elif msg_class is WatchUser and not msg.userexists:
                self._user_addresses.pop(msg.user, None)

        elif msg_class is Relogged:
            self._manual_server_disconnect = True
            self._server_relogged = True

        elif msg_class is PossibleParents:
            # Server sent a list of 10 potential parents, whose purpose is to forward us search requests.
            # We attempt to connect to them all at once, since connection errors are fairly common.

            self._potential_parents = msg.list
            log.add_conn("Server sent us a list of %s possible parents", len(msg.list))

            if self._parent_conn is None and self._potential_parents:
                for username, addr in self._potential_parents.items():
                    log.add_conn("Attempting parent connection to user %s", username)
                    self._initiate_connection_to_peer(username, ConnectionType.DISTRIBUTED, in_address=addr)

        elif msg_class is ParentMinSpeed:
            self._distrib_parent_min_speed = msg.speed
            log.add_conn("Received minimum distributed parent speed %s from the server", msg.speed)
            self._update_maximum_distributed_children()

        elif msg_class is ParentSpeedRatio:
            self._distrib_parent_speed_ratio = msg.ratio
            log.add_conn("Received distributed parent speed ratio %s from the server", msg.ratio)
            self._update_maximum_distributed_children()

        elif msg_class is ResetDistributed:
            log.add_conn("Received a reset request for distributed network")

            if self._parent_conn is not None:
                self._close_connection(self._parent_conn)

            for child_conn in self._child_peers.copy().values():
                self._close_connection(child_conn)

            self._send_have_no_parent()

        self._emit_network_message_event(msg)
        return True

    def _process_server_input(self, conn):
        """Reads messages from the input buffer of a server connection."""

        in_buffer = conn.in_buffer
        buffer_len = len(in_buffer)
        msg_content_offset = 8
        idx = 0

        # Server messages are 8 bytes or greater in length
        while buffer_len >= msg_content_offset:
            msg_size, msg_type = DOUBLE_UINT32_UNPACK(in_buffer, idx)

            if msg_size > self.MAX_INCOMING_MESSAGE_SIZE:
                log.add_conn("Received message larger than maximum size %s from server. "
                             "Closing connection.", self.MAX_INCOMING_MESSAGE_SIZE)
                self._manual_server_disconnect = True
                self._close_connection(conn)
                return

            msg_size_total = msg_size + 4

            if msg_size_total > buffer_len:
                # Buffer is being filled
                break

            # Unpack server messages
            if msg_type in SERVER_MESSAGE_CLASSES:
                if not self._process_server_message(
                    msg_type, msg_size, in_buffer, idx + msg_content_offset, idx + msg_size_total
                ):
                    self._manual_server_disconnect = True
                    self._close_connection(conn)
                    return
            else:
                msg_content = in_buffer[idx + msg_content_offset:idx + min(50, msg_size_total)]
                log.add_debug("Server message type %s size %s contents %s unknown",
                              (msg_type, msg_size, msg_content))

            idx += msg_size_total
            buffer_len -= msg_size_total

        if idx:
            del in_buffer[:idx]

    def _process_server_output(self, msg):

        msg_content = self._pack_network_message(msg)

        if msg_content is None:
            return

        msg_class = msg.__class__

        if msg_class is WatchUser and msg.user not in self._user_addresses:
            # Only cache IP address of watched users, otherwise we won't know if
            # a user reconnects and changes their IP address.
            self._user_addresses[msg.user] = None

        elif msg_class is UnwatchUser and msg.user != self._server_username:
            self._user_addresses.pop(msg.user, None)

        conn = self._server_conn
        out_buffer = conn.out_buffer

        out_buffer += msg.pack_uint32(len(msg_content) + 4)
        out_buffer += msg.pack_uint32(SERVER_MESSAGE_CODES[msg_class])
        out_buffer += msg_content

        self._modify_connection_events(conn, selectors.EVENT_READ | selectors.EVENT_WRITE)

    def _server_disconnect(self):
        """We're disconnecting from the server, clean up."""

        self._server_conn = None
        self._should_process_queue = False
        self._interface_name = self._interface_address = None
        self._local_ip_address = ""

        self._close_listen_socket()

        if self._portmapper is not None:
            self._portmapper.remove_port_mapping(blocking=True)
            self._portmapper.set_port(port=None, local_ip_address=None)
            self._portmapper = None

        self._parent_conn = None
        self._potential_parents.clear()
        self._branch_level = 0
        self._branch_root = None
        self._is_server_parent = False
        self._distrib_parent_min_speed = 0
        self._distrib_parent_speed_ratio = 1
        self._max_distrib_children = 0
        self._upload_speed = 0
        self._user_addresses.clear()

        self._check_indirect_request_timeouts(expire_all=True)

        for conn in self._conns.copy().values():
            self._close_connection(conn)

        self._message_queue.clear()
        self._pending_peer_conns.clear()
        self._pending_init_msgs.clear()
        self._username_init_msgs.clear()

        # Reset connection stats
        events.emit_main_thread("set-connection-stats")

        if not self._server_address:
            # We didn't successfully establish a connection to the server
            return

        ip_address, port = self._server_address

        log.add(
            _("Disconnected from server %(host)s:%(port)s"), {
                "host": ip_address,
                "port": port
            })

        if self._server_relogged:
            log.add(_("Someone logged in to your Soulseek account elsewhere"))
            self._server_relogged = False

        if not self._manual_server_disconnect:
            self._set_server_timer(use_fixed_timeout=self._manual_server_reconnect)

        self._server_address = None
        self._server_username = None

        events.emit_main_thread(
            "server-disconnect",
            ServerDisconnect(manual_disconnect=self._manual_server_disconnect)
        )

    def _send_message_to_server(self, msg):
        self._process_outgoing_messages([msg])

    # Peer Init #

    def _process_peer_init_message(self, conn, msg_type, msg_size, in_buffer, start_offset, end_offset):

        msg_class = PEER_INIT_MESSAGE_CLASSES[msg_type]
        msg = self._unpack_network_message(
            msg_class,
            memoryview(in_buffer)[start_offset:end_offset],
            msg_size,
            conn_type="peer init",
            sock=conn.sock
        )

        if msg is None:
            return None

        if msg_class is PierceFireWall:
            token = msg.token
            log.add_conn("Received indirect connection response (PierceFireWall) with token "
                         "%s, address %s", (token, conn.addr))

            log.add_conn("Number of stored peer init message tokens: %s", len(self._token_init_msgs))

            if token not in self._token_init_msgs:
                log.add_conn("Indirect connection attempt with token %s previously expired, "
                             "closing connection", token)
                return None

            init, _request_time = self._token_init_msgs.pop(token)
            previous_sock = init.sock
            is_direct_conn_in_progress = (
                previous_sock is not None and not self._conns[previous_sock].is_established
            )

            log.add_conn("Indirect connection to user %s with token %s established",
                         (init.target_user, token))

            if previous_sock is None or is_direct_conn_in_progress:
                init.sock = conn.sock
                log.add_conn("Using as primary connection, since no direct connection is established")
            else:
                # We already have a direct connection, but some clients may send a message over
                # the indirect connection. Keep it open.
                log.add_conn("Direct connection was already established, keeping it as primary connection")

            if is_direct_conn_in_progress:
                log.add_conn("Stopping direct connection attempt to user %s", init.target_user)
                self._close_connection(self._conns[previous_sock])

        elif msg_class is PeerInit:
            username = msg.target_user
            conn_type = msg.conn_type
            addr = conn.addr

            log.add_conn("Received incoming direct connection of type %s from user "
                         "%s, address %s", (conn_type, username, addr))

            if conn_type not in self.ALLOWED_PEER_CONN_TYPES:
                log.add_conn("Unknown connection type %s", conn_type)
                return None

            init = msg
            self._replace_existing_connection(init)

        self._emit_network_message_event(msg)
        return init

    def _process_peer_init_input(self, conn):
        """Reads peer init messages from the input buffer of a peer connection."""

        init = None
        in_buffer = conn.in_buffer
        buffer_len = len(in_buffer)
        msg_content_offset = 5
        idx = 0

        # Peer init messages are 5 bytes or greater in length
        while buffer_len >= msg_content_offset and init is None:
            msg_size, = UINT32_UNPACK(in_buffer, idx)

            if msg_size > self.MAX_INCOMING_MESSAGE_SIZE:
                log.add_conn("Received message larger than maximum size %s from peer %s. "
                             "Closing connection.", (self.MAX_INCOMING_MESSAGE_SIZE, conn.addr))
                break

            msg_size_total = msg_size + 4

            if msg_size_total > buffer_len:
                # Buffer is being filled
                conn.has_post_init_activity = True
                break

            # Unpack peer init messages
            msg_type = in_buffer[idx + 4]

            if msg_type in PEER_INIT_MESSAGE_CLASSES:
                init = self._process_peer_init_message(
                    conn, msg_type, msg_size, in_buffer, idx + msg_content_offset, idx + msg_size_total)
            else:
                msg_content = in_buffer[idx + msg_content_offset:idx + min(50, msg_size_total)]
                log.add_debug("Peer init message type %s size %s contents %s unknown",
                              (msg_type, msg_size, msg_content))

            if init is None:
                break

            idx += msg_size_total
            buffer_len -= msg_size_total

        if init is None:
            self._close_connection(conn)
            return None

        if idx:
            del in_buffer[:idx]

        conn.init = init

        self._add_init_message(init)
        self._process_conn_messages(init)
        self._accept_child_peer_connection(conn)
        return init

    def _process_peer_init_output(self, msg):

        # Pack peer init messages
        conn = self._conns[msg.sock]
        msg_content = self._pack_network_message(msg)

        if msg_content is None:
            return

        out_buffer = conn.out_buffer

        out_buffer += msg.pack_uint32(len(msg_content) + 1)
        out_buffer += msg.pack_uint8(PEER_INIT_MESSAGE_CODES[msg.__class__])
        out_buffer += msg_content

        self._modify_connection_events(conn, selectors.EVENT_READ | selectors.EVENT_WRITE)

    # Peer Connection #

    def _accept_incoming_peer_connections(self):

        while self._num_sockets < self.MAX_SOCKETS:
            incoming_sock = None

            try:
                incoming_sock, incoming_addr = self._listen_socket.accept()
                incoming_sock.setblocking(False)
                incoming_sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

            except OSError as error:
                if error.errno == errno.EWOULDBLOCK:
                    # No more incoming connections
                    break

                log.add_conn("Incoming connection failed: %s", error)

                if incoming_sock is not None:
                    self._close_socket(incoming_sock)

                continue

            io_events = selectors.EVENT_READ

            conn = self._conns[incoming_sock] = PeerConnection(
                sock=incoming_sock, addr=incoming_addr, io_events=io_events
            )
            self._num_sockets += 1

            # Event flags are modified to include 'write' in subsequent loops, if necessary.
            # Don't do it here, otherwise connections may break.
            self._selector.register(incoming_sock, io_events)
            conn.is_established = True

            log.add_conn("Incoming connection from address %s", (incoming_addr,))

    def _init_peer_connection(self, addr, init, response_token=None):

        if self._num_sockets >= self.MAX_SOCKETS:
            # Connection limit reached, re-queue
            self._pending_peer_conns[addr] = init
            return

        request_token = None
        _ip_address, port = addr
        self._pending_peer_conns.pop(addr, None)

        if response_token is None:
            # No token provided, we're not responding to an indirect connection request.
            # Request indirect connection from our end in case the user's port is closed.
            request_token = self._connect_to_peer_indirect(init)

        if port <= 0 or port > 65535:
            log.add_conn("Skipping direct connection attempt of type %s to user %s "
                         "due to invalid address %s", (init.conn_type, init.target_user, addr))
            return

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        io_events = selectors.EVENT_READ | selectors.EVENT_WRITE
        conn = PeerConnection(
            sock=sock, addr=addr, io_events=io_events,
            init=init, request_token=request_token, response_token=response_token
        )

        sock.setblocking(False)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        try:
            self._bind_socket_interface(sock)
            sock.connect_ex(addr)

        except OSError as error:
            self._connect_error(error, conn)
            self._close_socket(sock)
            return

        init.sock = sock
        self._conns[sock] = conn
        self._selector.register(sock, io_events)
        self._num_sockets += 1

    def _process_peer_input(self, conn):
        """Reads messages from the input buffer of a 'P' connection."""

        in_buffer = conn.in_buffer
        buffer_len = len(in_buffer)
        msg_content_offset = 8
        idx = 0
        search_result_received = False

        # Peer messages are 8 bytes or greater in length
        while buffer_len >= msg_content_offset:
            msg_size, msg_type = DOUBLE_UINT32_UNPACK(in_buffer, idx)

            if msg_size > self.MAX_INCOMING_MESSAGE_SIZE:
                log.add_conn("Received message larger than maximum size %s from user %s. "
                             "Closing connection.", (self.MAX_INCOMING_MESSAGE_SIZE, conn.init.target_user))
                self._close_connection(conn)
                return

            msg_size_total = msg_size + 4
            msg_class = None

            if msg_type in PEER_MESSAGE_CLASSES:
                msg_class = PEER_MESSAGE_CLASSES[msg_type]

            # Send progress to the main thread
            if msg_class is SharedFileListResponse:
                events.emit_main_thread(
                    "shared-file-list-progress", conn.init.target_user, conn.sock, buffer_len, msg_size_total)

            elif msg_class is UserInfoResponse:
                events.emit_main_thread(
                    "user-info-progress", conn.init.target_user, conn.sock, buffer_len, msg_size_total)

            if msg_size_total > buffer_len:
                # Buffer is being filled
                break

            # Unpack peer messages
            if msg_class:
                msg = self._unpack_network_message(
                    msg_class,
                    memoryview(in_buffer)[idx + msg_content_offset:idx + msg_size_total],
                    msg_size,
                    conn_type="peer",
                    sock=conn.sock,
                    addr=conn.addr,
                    username=conn.init.target_user
                )

                if msg_class is FileSearchResponse:
                    search_result_received = True

                self._emit_network_message_event(msg)
            else:
                msg_content = in_buffer[idx + msg_content_offset:idx + min(50, msg_size_total)]
                log.add_debug("Peer message type %s size %s contents %s unknown, from user: %s, address %s",
                              (msg_type, msg_size, msg_content, conn.init.target_user, conn.addr))

            idx += msg_size_total
            buffer_len -= msg_size_total

        if idx:
            del in_buffer[:idx]
            conn.has_post_init_activity = True

        if search_result_received and not self._is_connection_still_active(conn):
            # Forcibly close peer connection. Only used after receiving a search result,
            # as we need to get rid of peer connections before they pile up.

            self._close_connection(conn)

    def _process_peer_output(self, msg):

        # Pack peer messages
        msg_content = self._pack_network_message(msg)

        if msg_content is None:
            return

        conn = self._conns[msg.sock]
        out_buffer = conn.out_buffer

        out_buffer += msg.pack_uint32(len(msg_content) + 4)
        out_buffer += msg.pack_uint32(PEER_MESSAGE_CODES[msg.__class__])
        out_buffer += msg_content

        conn.has_post_init_activity = True
        self._modify_connection_events(conn, selectors.EVENT_READ | selectors.EVENT_WRITE)

    # File Connection #

    def _calc_upload_limit(self, limit_disabled=False, limit_per_transfer=False):

        limit = self._upload_limit
        loop_limit = 1024  # 1 KB/s is the minimum upload speed per transfer

        if limit_disabled or limit < loop_limit:
            self._upload_limit_split = 0
            return

        if not limit_per_transfer and self._total_uploads > 1:
            limit //= self._total_uploads

        self._upload_limit_split = int(limit)

    def _calc_upload_limit_by_transfer(self):
        return self._calc_upload_limit(limit_per_transfer=True)

    def _calc_upload_limit_none(self):
        return self._calc_upload_limit(limit_disabled=True)

    def _calc_download_limit(self):

        limit = self._download_limit
        loop_limit = 1024  # 1 KB/s is the minimum download speed per transfer

        if limit < loop_limit:
            # Download limit disabled
            self._download_limit_split = 0
            return

        if self._total_downloads > 1:
            limit //= self._total_downloads

        self._download_limit_split = int(limit)

    def _process_file_init_message(self, conn, in_buffer):

        msg_size = idx = 4
        msg = self._unpack_network_message(
            FileTransferInit,
            memoryview(in_buffer)[:msg_size],
            msg_size,
            conn_type="file",
            sock=conn.sock,
            username=conn.init.target_user
        )

        if msg is not None and msg.token is not None:
            self._file_init_msgs[conn] = msg
            self._emit_network_message_event(msg)

        return idx

    def _process_file_offset_message(self, conn, in_buffer):

        file_upload = self._file_upload_msgs[conn]

        if file_upload.offset is not None:
            # No more incoming messages on this connection after receiving the
            # file offset. If peer sends something anyway, clear it.
            return len(in_buffer)

        msg_size = idx = 8
        msg = self._unpack_network_message(
            FileOffset,
            memoryview(in_buffer)[:msg_size],
            msg_size,
            conn_type="file",
            sock=conn.sock,
            username=conn.init.target_user
        )

        if msg is None or msg.offset is None:
            return idx

        file_upload.offset = msg.offset

        events.emit_main_thread(
            "file-upload-progress",
            username=conn.init.target_user, token=file_upload.token,
            offset=file_upload.offset, bytes_sent=file_upload.sentbytes
        )

        try:
            file_upload.file.seek(msg.offset)
            self._modify_connection_events(conn, selectors.EVENT_READ | selectors.EVENT_WRITE)

        except (OSError, ValueError) as error:
            events.emit_main_thread(
                "upload-file-error",
                username=conn.init.target_user, token=file_upload.token, error=error
            )
            self._close_connection(conn)
            return None

        return idx

    def _write_download_file(self, file_download, data, data_len):

        try:
            if not data:
                return

            file_download.speed += data_len
            self._total_download_bandwidth += data_len

            file_download.file.write(data)
            file_download.leftbytes -= data_len

        finally:
            # Release memoryview in case of critical error
            data = None

    def _process_download(self, conn, data, data_len):

        file_download = self._file_download_msgs[conn]
        idx = file_download.leftbytes

        try:
            if data_len > idx:
                self._write_download_file(file_download, memoryview(data)[:idx], idx)
            else:
                self._write_download_file(file_download, data, data_len)

        except (OSError, ValueError) as error:
            events.emit_main_thread(
                "download-file-error",
                username=conn.init.target_user, token=file_download.token, error=error
            )
            return False  # Close the connection

        # Download finished
        if file_download.leftbytes <= 0:
            events.emit_main_thread(
                "file-download-progress",
                username=conn.init.target_user, token=file_download.token,
                bytes_left=file_download.leftbytes
            )
            return False  # Close the connection

        return True

    def _process_upload(self, conn, num_sent_bytes, current_time):

        file_upload = self._file_upload_msgs[conn]

        if file_upload.offset is None:
            return True

        out_buffer = conn.out_buffer
        out_buffer_len = len(out_buffer)
        file_upload.sentbytes += num_sent_bytes
        total_read_bytes = file_upload.offset + file_upload.sentbytes + out_buffer_len
        size = file_upload.size

        try:
            if total_read_bytes < size:
                num_bytes_to_read = int(
                    (max(4096, num_sent_bytes * 1.25) / max(1, current_time - conn.last_active))
                    - out_buffer_len
                )
                if num_bytes_to_read > 0:
                    out_buffer += file_upload.file.read(num_bytes_to_read)
                    self._modify_connection_events(conn, selectors.EVENT_READ | selectors.EVENT_WRITE)

        except (OSError, ValueError) as error:
            events.emit_main_thread(
                "upload-file-error",
                username=conn.init.target_user, token=file_upload.token, error=error
            )
            return False  # Close the connection

        file_upload.speed += num_sent_bytes
        self._total_upload_bandwidth += num_sent_bytes

        # Upload finished
        if file_upload.offset + file_upload.sentbytes == size:
            events.emit_main_thread(
                "file-upload-progress",
                username=conn.init.target_user, token=file_upload.token,
                offset=file_upload.offset, bytes_sent=file_upload.sentbytes
            )
        return True

    def _process_file_input(self, conn):
        """Reads file messages from the input buffer of a 'F' connection."""

        in_buffer = conn.in_buffer
        idx = 0

        if conn not in self._file_init_msgs:
            idx = self._process_file_init_message(conn, in_buffer)

        elif conn in self._file_upload_msgs:
            idx = self._process_file_offset_message(conn, in_buffer)

        if idx:
            del in_buffer[:idx]
            conn.has_post_init_activity = True

    def _process_file_output(self, msg):

        msg_class = msg.__class__

        # Pack file messages
        if msg_class is FileTransferInit:
            msg_content = self._pack_network_message(msg)

            if msg_content is None:
                return

            conn = self._conns[msg.sock]
            self._file_init_msgs[conn] = msg
            conn.out_buffer += msg_content

            self._emit_network_message_event(msg)

        elif msg_class is FileOffset:
            msg_content = self._pack_network_message(msg)

            if msg_content is None:
                return

            conn = self._conns[msg.sock]
            conn.out_buffer += msg_content

        conn.has_post_init_activity = True
        self._modify_connection_events(conn, selectors.EVENT_READ | selectors.EVENT_WRITE)

    # Distributed Connection #

    def _accept_child_peer_connection(self, conn):

        if conn.init.conn_type != ConnectionType.DISTRIBUTED:
            return

        username = conn.init.target_user

        if username == self._server_username:
            # We can't connect to ourselves
            return

        if username in self._potential_parents:
            # This is not a child peer, ignore
            return

        if self._parent_conn is None and not self._is_server_parent:
            # We have no parent user and the server hasn't sent search requests, no point
            # in accepting child peers
            log.add_conn("Rejecting distributed child peer connection from user %s, since we have no parent", username)
            self._close_connection(conn)
            return

        if username in self._child_peers:
            log.add_conn("Rejecting distributed child peer connection from user %s, since an existing connection "
                         "already exists", username)
            self._close_connection(conn)
            return

        if len(self._child_peers) >= self._max_distrib_children:
            log.add_conn("Rejecting distributed child peer connection from user %s, since child peer limit "
                         "of %s was reached", (username, self._max_distrib_children))
            self._close_connection(conn)
            return

        self._child_peers[username] = conn
        self._send_message_to_peer(username, DistribBranchLevel(self._branch_level))

        if self._parent_conn is not None:
            # Only sent when we're not the branch root
            self._send_message_to_peer(username, DistribBranchRoot(self._branch_root))

        log.add_conn("Adopting user %s as distributed child peer. Number of current child peers: %s",
                     (username, len(self._child_peers)))

        if len(self._child_peers) >= self._max_distrib_children:
            log.add_conn("Maximum number of distributed child peers reached (%s), "
                         "no longer accepting new connections", self._max_distrib_children)
            self._send_message_to_server(AcceptChildren(False))

    def _remove_child_peer_connection(self, username):

        self._child_peers.pop(username, None)

        if not self._should_process_queue:
            return

        if len(self._child_peers) == self._max_distrib_children - 1:
            log.add_conn("Available to accept a new distributed child peer")
            self._send_message_to_server(AcceptChildren(True))

        log.add_conn("Number of current child peers: %s", len(self._child_peers))

    def _send_message_to_child_peers(self, msg):

        msg_class = msg.__class__
        msg_attrs = [getattr(msg, s) for s in msg.__slots__]
        msgs = []

        for conn in self._child_peers.values():
            msg_child = msg_class(*msg_attrs)
            msg_child.sock = conn.sock
            msgs.append(msg_child)

        self._process_outgoing_messages(msgs)

    def _distribute_embedded_message(self, msg):
        """Distributes an embedded message from the server to our child
        peers."""

        if self._parent_conn is not None:
            # The server shouldn't send embedded messages while it's not our parent, but let's be safe
            return

        self._send_message_to_child_peers(DistribEmbeddedMessage(msg.distrib_code, msg.distrib_message))

        if self._is_server_parent:
            return

        self._is_server_parent = True

        if len(self._child_peers) < self._max_distrib_children:
            self._send_message_to_server(AcceptChildren(True))

        log.add_conn("Server is our parent, ready to distribute search requests as a branch root")

    def _verify_parent_connection(self, conn, msg_class):
        """Verify that a connection is our current parent connection."""

        if conn is not self._parent_conn:
            log.add_conn("Received a distributed message %s from user %s, who is not our parent. "
                         "Closing connection.", (msg_class, conn.init.target_user))
            return False

        return True

    def _send_have_no_parent(self):
        """Inform the server we have no parent.

        The server should either send us a PossibleParents message, or
        start sending us search requests.
        """

        if not self._should_process_queue:
            return

        # Note that we don't clear the previous list of possible parents here, since
        # it's possible the parent connection was closed immediately or superseded by
        # an indirect connection
        self._parent_conn = None
        self._branch_level = 0
        self._branch_root = self._server_username

        log.add_conn("We have no parent, requesting a new one")

        self._send_message_to_server(HaveNoParent(True))
        self._send_message_to_server(BranchRoot(self._branch_root))
        self._send_message_to_server(BranchLevel(self._branch_level))
        self._send_message_to_server(AcceptChildren(False))

    def _set_branch_root(self, username):
        """Inform the server and child peers of our branch root."""

        if not username:
            return

        if username == self._branch_root:
            return

        self._branch_root = username
        self._send_message_to_server(BranchRoot(username))
        self._send_message_to_child_peers(DistribBranchRoot(username))

        log.add_conn("Our branch root is user %s", username)

    def _update_maximum_distributed_children(self):

        prev_max_distrib_children = self._max_distrib_children
        num_child_peers = len(self._child_peers)

        if self._upload_speed >= self._distrib_parent_min_speed and self._distrib_parent_speed_ratio > 0:
            # Limit maximum distributed child peers to 10 for now due to socket limit concerns
            self._max_distrib_children = min(self._upload_speed // self._distrib_parent_speed_ratio // 100, 10)
        else:
            # Server does not allow us to accept distributed child peers
            self._max_distrib_children = 0

        log.add_conn("Distributed child peer limit updated, maximum connections: %s", self._max_distrib_children)

        if self._max_distrib_children <= num_child_peers < prev_max_distrib_children:
            log.add_conn("Our current number of distributed child peers (%s) reached the new limit, no longer "
                         "accepting new connections", num_child_peers)
            self._send_message_to_server(AcceptChildren(False))

    def _process_distrib_message(self, conn, msg_type, msg_size, in_buffer, start_offset, end_offset):

        msg_class = DISTRIBUTED_MESSAGE_CLASSES[msg_type]
        msg = self._unpack_network_message(
            msg_class,
            memoryview(in_buffer)[start_offset:end_offset],
            msg_size,
            conn_type="distrib",
            sock=conn.sock,
            username=conn.init.target_user
        )

        if msg is None:
            # Ignore unknown message and keep connection open
            return True

        if msg_class is DistribSearch:
            if not self._verify_parent_connection(conn, msg_class):
                return False

            self._send_message_to_child_peers(msg)

        elif msg_class is DistribEmbeddedMessage:
            if not self._verify_parent_connection(conn, msg_class):
                return False

            msg = self._unpack_embedded_message(msg)

            if msg is not None:
                self._send_message_to_child_peers(msg)

        elif msg_class is DistribBranchLevel:
            if msg.level < 0:
                # There are rare cases of parents sending a branch level value of -1,
                # presumably buggy clients
                log.add_conn("Received an invalid branch level value %s from user %s. "
                             "Closing connection.", (msg.level, msg.username))
                return False

            if self._parent_conn is None and msg.username in self._potential_parents:
                # We have a successful connection with a potential parent. Tell the server who
                # our parent is, and stop requesting new potential parents.
                self._parent_conn = conn
                self._branch_level = msg.level + 1
                self._is_server_parent = False

                self._send_message_to_server(HaveNoParent(False))
                self._send_message_to_server(BranchLevel(self._branch_level))

                if len(self._child_peers) < self._max_distrib_children:
                    self._send_message_to_server(AcceptChildren(True))

                self._send_message_to_child_peers(DistribBranchLevel(self._branch_level))
                self._child_peers.pop(msg.username, None)

                log.add_conn("Adopting user %s as parent", msg.username)
                log.add_conn("Our branch level is %s", self._branch_level)

                if self._branch_level == 1:
                    # Our current branch level is 1, our parent is a branch root
                    self._set_branch_root(msg.username)

            elif not self._verify_parent_connection(conn, msg_class):
                return False

            else:
                # Inform the server and child peers of our new branch level
                self._branch_level = msg.level + 1
                self._send_message_to_server(BranchLevel(self._branch_level))
                self._send_message_to_child_peers(DistribBranchLevel(self._branch_level))

                log.add_conn("Received a branch level update from our parent. Our new branch level is %s",
                             self._branch_level)

        elif msg_class is DistribBranchRoot:
            if not self._verify_parent_connection(conn, msg_class):
                return False

            self._set_branch_root(msg.root_username)

        self._emit_network_message_event(msg)
        return True

    def _process_distrib_input(self, conn):
        """Reads messages from the input buffer of a 'D' connection."""

        in_buffer = conn.in_buffer
        buffer_len = len(in_buffer)
        msg_content_offset = 5
        idx = 0

        # Distributed messages are 5 bytes or greater in length
        while buffer_len >= msg_content_offset:
            msg_size, = UINT32_UNPACK(in_buffer, idx)

            if msg_size > self.MAX_INCOMING_MESSAGE_SIZE:
                log.add_conn("Received message larger than maximum size %s from user %s. "
                             "Closing connection.", (self.MAX_INCOMING_MESSAGE_SIZE, conn.init.target_user))
                self._close_connection(conn)
                break

            msg_size_total = msg_size + 4

            if msg_size_total > buffer_len:
                # Buffer is being filled
                conn.has_post_init_activity = True
                break

            # Unpack distributed messages
            msg_type = in_buffer[idx + 4]

            if msg_type in DISTRIBUTED_MESSAGE_CLASSES:
                if not self._process_distrib_message(
                    conn, msg_type, msg_size, in_buffer, idx + msg_content_offset, idx + msg_size_total
                ):
                    self._close_connection(conn)
                    return
            else:
                msg_content = in_buffer[idx + msg_content_offset:idx + min(50, msg_size_total)]
                log.add_debug("Distrib message type %s size %s contents %s unknown",
                              (msg_type, msg_size, msg_content))

            idx += msg_size_total
            buffer_len -= msg_size_total

        if idx:
            del in_buffer[:idx]
            conn.has_post_init_activity = True

    def _process_distrib_output(self, msg):

        # Pack distributed messages
        msg_content = self._pack_network_message(msg)

        if msg_content is None:
            return

        conn = self._conns[msg.sock]
        out_buffer = conn.out_buffer

        out_buffer += msg.pack_uint32(len(msg_content) + 1)
        out_buffer += msg.pack_uint8(DISTRIBUTED_MESSAGE_CODES[msg.__class__])
        out_buffer += msg_content

        conn.has_post_init_activity = True
        self._modify_connection_events(conn, selectors.EVENT_READ | selectors.EVENT_WRITE)

    # Internal Messages #

    def _process_internal_messages(self, msg):

        msg_class = msg.__class__

        if msg_class is CloseConnection:
            self._close_connection(self._conns.get(msg.sock))

        elif msg_class is ServerConnect:
            self._server_connect(msg)

        elif msg_class is ServerDisconnect:
            self._manual_server_disconnect = True
            self._close_connection(self._server_conn)

        elif msg_class is ServerReconnect:
            self._manual_server_reconnect = True
            self._close_connection(self._server_conn)

        elif msg_class is DownloadFile:
            conn = self._conns.get(msg.sock)

            if conn is not None:
                self._file_download_msgs[conn] = msg

                self._total_downloads += 1
                self._calc_download_limit()
                self._process_conn_incoming_messages(conn)

        elif msg_class is UploadFile:
            conn = self._conns.get(msg.sock)

            if conn is not None:
                self._file_upload_msgs[conn] = msg

                self._total_uploads += 1
                self._calc_upload_limit_function()
                self._process_conn_incoming_messages(conn)

        elif msg_class is SetDownloadLimit:
            self._download_limit = msg.limit * 1024
            self._calc_download_limit()

        elif msg_class is SetUploadLimit:
            if msg.limit > 0:
                if msg.limitby:
                    self._calc_upload_limit_function = self._calc_upload_limit
                else:
                    self._calc_upload_limit_function = self._calc_upload_limit_by_transfer
            else:
                self._calc_upload_limit_function = self._calc_upload_limit_none

            self._upload_limit = msg.limit * 1024
            self._calc_upload_limit_function()

        elif msg_class is EmitNetworkMessageEvents:
            for network_msg in msg.msgs:
                self._emit_network_message_event(network_msg)

    # Input/Output #

    def _process_ready_input_socket(self, sock, current_time):

        if sock in self._conns:
            conn = self._conns[sock]
        else:
            # Unknown connection
            return

        if (self._download_limit_split
                and conn in self._conns_downloaded
                and self._conns_downloaded[conn] >= self._download_limit_split):
            return

        conn_error = None

        try:
            if self._read_data(conn, current_time):
                self._process_conn_incoming_messages(conn)
                return

        except OSError as error:
            log.add_conn("Cannot read data from connection %s, closing connection. "
                         "Error: %s", (conn.addr, error))
            conn_error = error

        if not conn.is_established:
            if conn_error is None:
                # No error when connection shuts down gracefully (recv() returns
                # 0 bytes), but we need to display one anyway. Is this is the best fit?
                conn_error = self.ERROR_NOT_CONNECTED

            self._connect_error(conn_error, conn)

        self._close_connection(conn)

    def _process_ready_output_socket(self, sock, current_time):

        if sock in self._conns:
            conn = self._conns[sock]
        else:
            # Unknown connection
            return

        if not conn.is_established:
            if conn is self._server_conn:
                self._establish_outgoing_server_connection(conn)
            else:
                self._establish_outgoing_peer_connection(conn)

            if sock not in self._conns:
                # Connection was closed while being established
                return

        if (self._upload_limit_split
                and conn in self._conns_uploaded
                and self._conns_uploaded[conn] >= self._upload_limit_split):
            return

        try:
            if self._write_data(conn, current_time):
                return

        except (OSError, ValueError) as error:
            log.add_conn("Cannot write data to connection %s, closing connection. Error: %s",
                         (conn.addr, error))

        self._close_connection(conn)

    def _process_ready_sockets(self, current_time):

        if self._listen_socket is None:
            # We can't call select() when no sockets are registered (WinError 10022)
            return

        for key, io_events in self._selector.select(timeout=self.SLEEP_MAX_IDLE):
            sock = key.fileobj

            if io_events & selectors.EVENT_READ:
                if sock is self._listen_socket:
                    self._accept_incoming_peer_connections()
                    continue

                self._process_ready_input_socket(sock, current_time)

            if io_events & selectors.EVENT_WRITE:
                self._process_ready_output_socket(sock, current_time)

    def _process_conn_incoming_messages(self, conn):

        if not conn.in_buffer:
            return

        if conn is self._server_conn:
            self._process_server_input(conn)
            return

        init = conn.init

        if init is None:
            conn.init = init = self._process_peer_init_input(conn)

            if init is None or not conn.in_buffer:
                return

        if init.conn_type == ConnectionType.PEER:
            self._process_peer_input(conn)

        elif init.conn_type == ConnectionType.FILE:
            self._process_file_input(conn)

        elif init.conn_type == ConnectionType.DISTRIBUTED:
            self._process_distrib_input(conn)

        if conn.sock is not None and init.sock is not conn.sock:
            log.add_conn("Received message on secondary connection of type %s to user %s, "
                         "promoting to primary connection", (init.conn_type, init.target_user))
            init.sock = conn.sock

    def _process_outgoing_messages(self, msgs):

        for msg in msgs:
            if not self._should_process_queue:
                return

            msg_type = msg.msg_type
            process_func = None

            if msg_type == MessageType.INIT:
                process_func = self._process_peer_init_output
                sock = msg.sock

            elif msg_type == MessageType.INTERNAL:
                process_func = self._process_internal_messages
                sock = None

            elif msg_type == MessageType.PEER:
                process_func = self._process_peer_output
                sock = msg.sock

                if sock is None:
                    self._send_message_to_peer(msg.username, msg)
                    continue

            elif msg_type == MessageType.DISTRIBUTED:
                process_func = self._process_distrib_output
                sock = msg.sock

            elif msg_type == MessageType.FILE:
                process_func = self._process_file_output
                sock = msg.sock

                if sock is None:
                    self._send_message_to_peer(msg.username, msg)
                    continue

            elif msg_type == MessageType.SERVER:
                process_func = self._process_server_output
                sock = self._server_conn.sock

            log.add_msg_contents(msg, is_outgoing=True)

            if sock is not None and sock not in self._conns:
                log.add_conn("Cannot send the message over the closed connection: %s %s",
                             (msg.__class__, msg))
                continue

            if process_func is not None:
                process_func(msg)

    def _process_queue_messages(self):

        if not self._message_queue:
            return

        msgs = []

        while self._message_queue:
            msgs.append(self._message_queue.popleft())

        self._process_outgoing_messages(msgs)

    def _read_data(self, conn, current_time):

        sock = conn.sock
        current_recv_size = conn.recv_size
        is_file_download = (conn in self._file_download_msgs)
        use_download_limit = (self._download_limit_split and is_file_download)

        if use_download_limit:
            download_limit = (self._download_limit_split - self._conns_downloaded[conn])

            if current_recv_size > download_limit:  # pylint: disable=consider-using-min-builtin
                current_recv_size = download_limit

        data = sock.recv(current_recv_size)
        data_len = len(data)

        if not data:
            return False  # Close the connection

        # An intermediate buffer is useless when downloading a file. Write to the
        # file immediately, and let the OS handle buffering when necessary.
        if not is_file_download:
            conn.in_buffer += data

        elif not self._process_download(conn, data, data_len):
            return False  # Close the connection

        if use_download_limit:
            self._conns_downloaded[conn] += data_len

        # Grow or shrink recv buffer depending on how much data we're receiving
        elif data_len >= current_recv_size // 2:
            conn.recv_size *= 2

        elif data_len <= current_recv_size // 6:
            conn.recv_size //= 2

        conn.last_active = current_time
        return True

    def _write_data(self, conn, current_time):

        sock = conn.sock
        out_buffer = conn.out_buffer
        is_file_upload = (conn in self._file_upload_msgs)

        if is_file_upload and self._upload_limit_split:
            limit = (self._upload_limit_split - self._conns_uploaded[conn])

            if len(out_buffer) > limit:
                num_bytes_sent = sock.send(memoryview(out_buffer)[:limit])
            else:
                num_bytes_sent = sock.send(out_buffer)

            self._conns_uploaded[conn] += num_bytes_sent
        else:
            num_bytes_sent = sock.send(out_buffer)

        del out_buffer[:num_bytes_sent]

        if is_file_upload and not self._process_upload(conn, num_bytes_sent, current_time):
            return False  # Close the connection

        if not out_buffer:
            # Nothing else to send, stop watching connection for writes
            self._modify_connection_events(conn, selectors.EVENT_READ)

        conn.last_active = current_time
        return True

    # Networking Loop #

    def _loop(self):

        while not self._want_abort:
            current_time = time.monotonic()

            if (current_time - self._last_cycle_time) >= 1:
                self._check_connections(current_time)
                self._check_indirect_request_timeouts(current_time)

                events.emit_main_thread(
                    "set-connection-stats",
                    total_conns=self._num_sockets,
                    download_bandwidth=self._total_download_bandwidth,
                    upload_bandwidth=self._total_upload_bandwidth
                )

                self._conns_downloaded.clear()
                self._conns_uploaded.clear()

                self._total_download_bandwidth = 0
                self._total_upload_bandwidth = 0

                self._last_cycle_time = current_time

            if not self._should_process_queue:
                if self._server_timeout_time and (self._server_timeout_time - current_time) <= 0:
                    self._server_timeout_time = None
                    events.emit_main_thread(
                        "server-reconnect",
                        ServerReconnect(manual_reconnect=self._manual_server_reconnect)
                    )

                time.sleep(self.SLEEP_MAX_IDLE + self.SLEEP_MIN_IDLE)
                continue

            # Process queue messages
            self._process_queue_messages()

            # Check which connections are ready to send/receive data
            self._process_ready_sockets(current_time)

            # Don't exhaust the CPU
            time.sleep(self.SLEEP_MIN_IDLE)

    def run(self):

        events.emit_main_thread("set-connection-stats")

        # Watch sockets for I/0 readiness with the selectors module. Only call register() after a socket
        # is bound, otherwise watching the socket not guaranteed to work (breaks on OpenBSD at least)
        self._selector = selectors.DefaultSelector()

        try:
            self._loop()

        finally:
            # Networking thread aborted
            self._manual_server_disconnect = True
            self._close_connection(self._server_conn)
            self._selector.close()

            # We're ready to quit
            events.emit_main_thread("quit")



# COPYRIGHT (C) 2022-2025 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

from gi.repository import Gtk

from pynicotine.gtkgui.application import GTK_API_VERSION
from pynicotine.gtkgui.application import GTK_MINOR_VERSION
from pynicotine.gtkgui.application import GTK_MICRO_VERSION
from pynicotine.gtkgui.application import LIBADWAITA_API_VERSION


class Window:

    DWMWA_USE_IMMERSIVE_DARK_MODE_BEFORE_20H1 = 19
    DWMWA_USE_IMMERSIVE_DARK_MODE = 20

    active_dialogs = []  # Class variable keeping dialog objects alive
    activation_token = None

    def __init__(self, widget):

        self.widget = widget
        self._dark_mode_handler = None
        self._text_widget = None

        if GTK_API_VERSION == 3:
            return

        self.widget.connect("notify::focus-widget", self._on_focus_widget_changed)

        if sys.platform == "win32":
            widget.connect("realize", self._on_realize_win32)

            # Use dark window controls on Windows when requested
            if LIBADWAITA_API_VERSION:
                from gi.repository import Adw  # pylint: disable=no-name-in-module
                self._dark_mode_handler = Adw.StyleManager.get_default().connect(
                    "notify::dark", self._on_dark_mode_win32
                )

        elif os.environ.get("GDK_BACKEND") == "broadway":
            # Workaround for GTK 4 bug where broadwayd uses a lot of CPU after hiding window
            self.widget.connect("hide", self._on_hide_broadway)

    def _on_focus_widget_changed(self, *_args):

        focus_widget = self.widget.get_focus()

        if focus_widget is None:
            return

        popover = focus_widget.get_ancestor(Gtk.Popover)

        # Workaround for GTK 4 bug where broadwayd uses a lot of CPU after hiding popover
        if popover is not None and os.environ.get("GDK_BACKEND") == "broadway":
            popover.hide_handler_broadway = popover.connect_after("hide", self._on_popover_hide_broadway)

        if not (4, 16, 0) <= (GTK_API_VERSION, GTK_MINOR_VERSION, GTK_MICRO_VERSION) <= (4, 16, 5):
            return

        # Workaround for GTK 4.16 bug where text is replaced when inserting emoji
        if not isinstance(popover, Gtk.EmojiChooser):
            self._text_widget = focus_widget if isinstance(focus_widget, Gtk.Text) else None
            return

        if self._text_widget is None:
            return

        popover.old_text = self._text_widget.get_text()
        popover.old_pos = self._text_widget.get_position()

        popover.picked_handler = popover.connect("emoji-picked", self._on_emoji_chooser_picked)
        popover.hide_handler = popover.connect("hide", self._on_emoji_chooser_hide)

    def _on_emoji_chooser_picked(self, chooser, emoji):

        old_text = chooser.old_text
        old_pos = chooser.old_pos
        new_text = old_text[:old_pos] + emoji + old_text[old_pos:]

        self._text_widget.set_text(new_text)
        self._text_widget.set_position(old_pos + len(emoji))

    def _on_emoji_chooser_hide(self, chooser):

        if chooser.picked_handler is not None:
            chooser.disconnect(chooser.picked_handler)
            chooser.picked_handler = None

        if chooser.hide_handler is not None:
            chooser.disconnect(chooser.hide_handler)
            chooser.hide_handler = None

    def _on_popover_hide_broadway(self, popover):

        if popover.hide_handler_broadway is not None:
            popover.disconnect(popover.hide_handler_broadway)
            popover.hide_handler_broadway = None

        popover.unrealize()

    def _on_realize_win32(self, *_args):

        from ctypes import windll

        # Don't overlap taskbar when auto-hidden
        h_wnd = self.get_surface().get_handle()
        windll.user32.SetPropW(h_wnd, "NonRudeHWND", True)

        # Set dark window controls
        if LIBADWAITA_API_VERSION:
            from gi.repository import Adw  # pylint: disable=no-name-in-module
            self._on_dark_mode_win32(Adw.StyleManager.get_default())

    def _on_dark_mode_win32(self, style_manager, *_args):

        surface = self.get_surface()

        if surface is None:
            return

        h_wnd = surface.get_handle()

        if h_wnd is None:
            return

        from ctypes import byref, c_int, sizeof, windll

        value = c_int(int(style_manager.get_dark()))

        if not windll.dwmapi.DwmSetWindowAttribute(
            h_wnd, self.DWMWA_USE_IMMERSIVE_DARK_MODE, byref(value), sizeof(value)
        ):
            windll.dwmapi.DwmSetWindowAttribute(
                h_wnd, self.DWMWA_USE_IMMERSIVE_DARK_MODE_BEFORE_20H1, byref(value), sizeof(value)
            )

    def _on_hide_broadway(self, *_args):
        self.widget.unrealize()

    def get_surface(self):

        if GTK_API_VERSION >= 4:
            return self.widget.get_surface()

        return self.widget.get_window()

    def get_width(self):

        if GTK_API_VERSION >= 4:
            return self.widget.get_width()

        width, _height = self.widget.get_size()
        return width

    def get_height(self):

        if GTK_API_VERSION >= 4:
            return self.widget.get_height()

        _width, height = self.widget.get_size()
        return height

    def get_position(self):

        if GTK_API_VERSION >= 4:
            return None

        return self.widget.get_position()

    def is_active(self):
        return self.widget.is_active()

    def is_maximized(self):
        return self.widget.is_maximized()

    def is_visible(self):
        return self.widget.get_visible()

    def set_title(self, title):
        self.widget.set_title(title)

    def maximize(self):
        self.widget.maximize()

    def unmaximize(self):
        self.widget.unmaximize()

    def present(self):

        if self.activation_token is not None:
            # Set XDG activation token if provided by tray icon
            self.widget.set_startup_id(self.activation_token)

        self.widget.present()

    def hide(self):
        self.widget.set_visible(False)

    def close(self, *_args):
        self.widget.close()

    def destroy(self):

        if self._dark_mode_handler is not None:
            from gi.repository import Adw  # pylint: disable=no-name-in-module
            Adw.StyleManager.get_default().disconnect(self._dark_mode_handler)

        self.widget.destroy()
        self.__dict__.clear()



# COPYRIGHT (C) 2020-2025 Nicotine+ Contributors
# COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
# COPYRIGHT (C) 2016 Mutnick <muhing@yahoo.com>
# COPYRIGHT (C) 2009-2011 quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2009 daelstorm <daelstorm@gmail.com>
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gc
import mmap
import os
import stat
import sys
import time

from collections import defaultdict
from collections import deque
from itertools import chain
from os import SEEK_END
from os import SEEK_SET
from pickle import HIGHEST_PROTOCOL
from pickle import dumps
from pickle import Unpickler
from pickle import UnpicklingError
from struct import Struct
from threading import Thread

from pynicotine import rename_process
from pynicotine.config import config
from pynicotine.core import core
from pynicotine.events import events
from pynicotine.external.tinytag import TinyTag
from pynicotine.logfacility import log
from pynicotine.slskmessages import FileListMessage
from pynicotine.slskmessages import FolderContentsResponse
from pynicotine.slskmessages import GetUserStats
from pynicotine.slskmessages import SharedFileListResponse
from pynicotine.slskmessages import SharedFoldersFiles
from pynicotine.utils import TRANSLATE_PUNCTUATION
from pynicotine.utils import UINT32_LIMIT
from pynicotine.utils import encode_path


class FileTypes:
    ARCHIVE = {
        "7z", "br", "bz2", "gz", "iso", "lz", "lzma", "rar", "tar", "tbz", "tbz2", "tgz", "xz", "zip", "zst"
    }
    AUDIO = {
        "aac", "ac3", "afc", "aifc", "aif", "aiff", "ape", "au", "bwav", "bwf", "dff", "dsd", "dsf", "dts", "flac",
        "gbs", "gym", "it", "m4a", "m4b", "mid", "midi", "mka", "mod", "mp1", "mp2", "mp3", "mp+", "mpc", "nsf", "nsfe",
        "ofr", "ofs", "oga", "ogg", "opus", "psf", "psf1", "psf2", "s3m", "sid", "spc", "spx", "ssf", "tak", "tta",
        "wav", "vgm", "vgz", "wma", "vqf", "wv", "xm"
    }
    EXECUTABLE = {
        "apk", "appimage", "bat", "deb", "dmg", "flatpak", "exe", "jar", "msi", "pkg", "rpm", "sh"
    }
    IMAGE = {
        "apng", "avif", "bmp", "gif", "heic", "heif", "ico", "jfif", "jp2", "jpg", "jpe", "jpeg", "jxl", "png", "psd",
        "raw", "svg", "svgz", "tif", "tiff", "webp"
    }
    DOCUMENT = {
        "doc", "docx", "epub", "mobi", "odp", "ods", "odt", "opf", "oxps", "pdf", "ppt", "pptx", "rtf", "xls", "xlsx",
        "xps"
    }
    TEXT = {
        "cue", "csv", "htm", "html", "m3u", "m3u8", "md5", "log", "lrc", "md", "mks", "nfo", "ps", "rst", "sfv",
        "sha1", "sha256", "srt", "txt"
    }
    VIDEO = {
        "3gp", "amv", "asf", "avi", "f4v", "flv", "m2ts", "m2v", "m4p", "m4v", "mov", "mp4", "mpe", "mpeg", "mpg",
        "mkv", "mts", "ogv", "ts", "vob", "webm", "wmv"
    }


class PermissionLevel:
    PUBLIC = "public"
    BUDDY = "buddy"
    TRUSTED = "trusted"
    BANNED = "banned"


class RestrictedUnpickler(Unpickler):
    """Don't allow code execution from pickles."""

    def find_class(self, module, name):
        # Forbid all globals
        raise UnpicklingError(f"global '{module}.{name}' is forbidden")


class DatabaseError(Exception):
    pass


class Database:
    """Custom key-value database format for Nicotine+ shares."""

    __slots__ = ("_value_offsets", "_file_handle", "_file_offset", "_overwrite")

    FILE_SIGNATURE = b"DBN+"
    VERSION = 3
    LENGTH_DATA_SIZE = 8
    PACK_LENGTHS = Struct("!II").pack
    UNPACK_LENGTHS = Struct("!II").unpack_from
    PICKLE_PROTOCOL = min(HIGHEST_PROTOCOL, 5)  # Use version 5 when available

    def __init__(self, file_path, overwrite=True):

        folder_path = os.path.dirname(file_path)
        mode = "ab+" if overwrite else "rb"

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

        if overwrite and os.path.exists(file_path):
            os.remove(file_path)

        self._value_offsets = self._load_value_offsets(file_path, mode)

        if overwrite:
            self._file_handle = open(file_path, mode)  # pylint: disable=consider-using-with
        else:
            with open(file_path, mode) as file_handle:
                self._file_handle = mmap.mmap(file_handle.fileno(), length=0, access=mmap.ACCESS_READ)

        self._file_offset = self._file_handle.seek(0, SEEK_END)
        self._overwrite = overwrite

    def _parse_content(self, content, total_size):

        value_offsets = {}
        file_signature_length = len(self.FILE_SIGNATURE)

        if content[:file_signature_length] != self.FILE_SIGNATURE:
            raise DatabaseError("Not a database file")

        if content[file_signature_length:file_signature_length + 1][0] != self.VERSION:
            raise DatabaseError("Incompatible version")

        current_offset = (file_signature_length + 1)

        while current_offset < total_size:
            key_offset = (current_offset + self.LENGTH_DATA_SIZE)
            key_length, value_length = self.UNPACK_LENGTHS(content[current_offset:key_offset])
            value_offset = (key_offset + key_length)
            key = content[key_offset:value_offset].tobytes().decode("utf-8")

            value_offsets[key] = value_offset
            current_offset = (value_offset + value_length)

        return value_offsets

    def _load_value_offsets(self, file_path, mode):

        value_offsets = {}

        with open(file_path, mode) as file_handle:  # pylint: disable=unspecified-encoding
            file_size = os.fstat(file_handle.fileno()).st_size

            if not file_size:
                file_handle.write(self.FILE_SIGNATURE)
                file_handle.write(bytes([self.VERSION]))
                return value_offsets

            file_handle.seek(0)

            with mmap.mmap(file_handle.fileno(), length=0, access=mmap.ACCESS_READ) as content:
                value_offsets = self._parse_content(memoryview(content), file_size)

        return value_offsets

    def __contains__(self, key):
        return key in self._value_offsets

    def __iter__(self):
        yield from self._value_offsets

    def __len__(self):
        return len(self._value_offsets)

    def __getitem__(self, key):

        value_offset = self._value_offsets[key]

        self._file_handle.seek(value_offset, SEEK_SET)
        return RestrictedUnpickler(self._file_handle).load()

    def __setitem__(self, key, value):

        encoded_key = key.encode("utf-8")
        pickled_value = dumps(value, protocol=self.PICKLE_PROTOCOL)

        key_length = len(encoded_key)
        length_data = self.PACK_LENGTHS(key_length, len(pickled_value))
        item_data = (length_data + encoded_key + pickled_value)

        self._file_handle.write(item_data)

        self._value_offsets[key] = self._file_offset + self.LENGTH_DATA_SIZE + key_length
        self._file_offset += len(item_data)

    def get(self, key, default=None):

        if key in self._value_offsets:
            return self[key]

        return default

    def update(self, obj):
        for key, value in obj.items():
            self[key] = value

    def close(self):

        if self._overwrite:
            os.fsync(self._file_handle)

        self._file_handle.close()


class ScannerState:
    INITIALIZED = "initialized"
    RESCANNING = "rescanning"
    FAILURE = "failure"


class ScannerLogMessage:
    __slots__ = ("msg", "msg_args")

    def __init__(self, msg, msg_args=None):
        self.msg = msg
        self.msg_args = msg_args


class Scanner:
    """Separate process responsible for building shares.

    It handles scanning of folders and files, as well as building
    databases and writing them to disk.
    """

    __slots__ = ("queue", "share_groups", "share_dbs", "share_db_paths", "init",
                 "rescan", "rebuild", "reveal_buddy_shares", "reveal_trusted_shares",
                 "files", "streams", "mtimes", "word_index", "processed_share_names",
                 "processed_share_paths", "current_file_index", "current_folder_count")

    HIDDEN_FOLDER_NAMES = {"@eaDir", "#recycle", "#snapshot"}

    def __init__(self, queue, share_groups, share_db_paths, init=False, rescan=True,
                 rebuild=False, reveal_buddy_shares=False, reveal_trusted_shares=False):

        self.queue = queue
        self.share_groups = share_groups
        self.share_dbs = {}
        self.share_db_paths = share_db_paths
        self.init = init
        self.rescan = rescan
        self.rebuild = rebuild
        self.reveal_buddy_shares = reveal_buddy_shares
        self.reveal_trusted_shares = reveal_trusted_shares
        self.files = {}
        self.streams = {}
        self.mtimes = {}
        self.word_index = defaultdict(list)
        self.processed_share_names = set()
        self.processed_share_paths = set()
        self.current_file_index = 0
        self.current_folder_count = 0

    def run(self):

        try:
            rename_process(b"nicotine-scan")

            if self.init:
                try:
                    self.create_compressed_shares()
                    self.create_file_path_index()

                except Exception:
                    # Failed to load shares or version is invalid, rebuild
                    self.rescan = self.rebuild = True

                self.queue.put(ScannerState.INITIALIZED)

            if self.rescan:
                self.queue.put(ScannerState.RESCANNING)
                self.queue.put(
                    ScannerLogMessage(_("Rebuilding shares…") if self.rebuild else _("Rescanning shares…"))
                )

                # Clear previous word index to prevent inconsistent state if the scanner fails
                self.set_shares(word_index={})

                # Scan shares
                for permission_level in (
                    PermissionLevel.PUBLIC,
                    PermissionLevel.BUDDY,
                    PermissionLevel.TRUSTED
                ):
                    self.rescan_dirs(permission_level)

                self.set_shares(word_index=self.word_index)
                self.word_index.clear()

                self.create_compressed_shares()
                self.create_file_path_index()

                self.queue.put(
                    ScannerLogMessage(
                        _("Rescan complete: %(num)s folders found"),
                        {"num": self.current_folder_count}
                    )
                )

        except Exception:
            from traceback import format_exc

            self.queue.put(
                ScannerLogMessage(
                    _("Serious error occurred while rescanning shares. If this problem persists, "
                      "delete %(dir)s/*.dbn and try again. If that doesn't help, please file a bug "
                      "report with this stack trace included: %(trace)s"), {
                        "dir": os.path.dirname(next(iter(self.share_db_paths.values()))),
                        "trace": "\n" + format_exc()
                    }
                )
            )
            self.queue.put(ScannerState.FAILURE)

        finally:
            Shares.close_shares(self.share_dbs)

    def create_compressed_shares_message(self, permission_level):
        """Create a message that will later contain a compressed list of our
        shares."""

        public_streams = self.share_dbs["public_streams"]
        buddy_streams = self.share_dbs["buddy_streams"]
        trusted_streams = self.share_dbs["trusted_streams"]

        if permission_level == PermissionLevel.PUBLIC and not self.reveal_buddy_shares:
            buddy_streams = None

        if permission_level in {PermissionLevel.PUBLIC, PermissionLevel.BUDDY} and not self.reveal_trusted_shares:
            trusted_streams = None

        compressed_shares = SharedFileListResponse(
            public_shares=public_streams, buddy_shares=buddy_streams, trusted_shares=trusted_streams,
            permission_level=permission_level
        )
        compressed_shares.make_network_message()
        compressed_shares.public_shares = compressed_shares.buddy_shares = compressed_shares.trusted_shares = None

        self.queue.put(compressed_shares)

    def create_compressed_shares(self):

        Shares.load_shares(
            self.share_dbs, self.share_db_paths, destinations={"public_streams", "buddy_streams", "trusted_streams"}
        )

        for permission_level in (PermissionLevel.PUBLIC, PermissionLevel.BUDDY, PermissionLevel.TRUSTED):
            self.create_compressed_shares_message(permission_level)

        Shares.close_shares(self.share_dbs)

    def create_file_path_index(self):

        Shares.load_shares(
            self.share_dbs, self.share_db_paths, destinations={"public_files", "buddy_files", "trusted_files"}
        )

        file_path_index = tuple(chain(
            self.share_dbs["public_files"],
            self.share_dbs["buddy_files"],
            self.share_dbs["trusted_files"]
        ))
        self.queue.put(file_path_index)

        Shares.close_shares(self.share_dbs)

    def real2virtual(self, real_path):

        real_path = real_path.replace("/", "\\")

        for shared_folder_paths in self.share_groups:
            for virtual_name, folder_path, *_unused in shared_folder_paths:
                folder_path = folder_path.replace("/", "\\")

                if real_path == folder_path:
                    return virtual_name

                # Use rstrip to remove trailing separator from root folders
                folder_path = folder_path.rstrip("\\") + "\\"

                if real_path.startswith(folder_path):
                    real_path_no_prefix = real_path[len(folder_path):]
                    virtual_path = f"{virtual_name}\\{real_path_no_prefix}"
                    return virtual_path

        raise ValueError(f"Cannot find virtual path for {real_path}")

    def set_shares(self, permission_level=None, files=None, streams=None, mtimes=None, word_index=None):

        for source, destination in (
            (files, "files"),
            (streams, "streams"),
            (mtimes, "mtimes"),
            (word_index, "words")
        ):
            if source is None:
                continue

            if destination != "words":
                destination = f"{permission_level}_{destination}"

            share_db = None

            try:
                share_db_path = self.share_db_paths[destination]
                share_db = Shares.create_db_file(share_db_path)
                share_db.update(source)

            finally:
                if share_db is not None:
                    share_db.close()

    def rescan_dirs(self, permission_level):

        shared_public_folders, shared_buddy_folders, shared_trusted_folders = self.share_groups

        if permission_level == PermissionLevel.TRUSTED:
            shared_folder_paths = sorted(shared_trusted_folders)

        elif permission_level == PermissionLevel.BUDDY:
            shared_folder_paths = sorted(shared_buddy_folders)

        else:
            shared_folder_paths = sorted(shared_public_folders)

        try:
            Shares.load_shares(
                self.share_dbs, self.share_db_paths,
                destinations={f"{permission_level}_files", f"{permission_level}_mtimes"})

        except Exception:
            # No previous share databases, rebuild
            self.rebuild = True

        old_files = self.share_dbs.get(f"{permission_level}_files")
        old_mtimes = self.share_dbs.get(f"{permission_level}_mtimes")

        for virtual_name, folder_path, *_unused in shared_folder_paths:
            if virtual_name in self.processed_share_names:
                # No duplicate names
                continue

            if folder_path in self.processed_share_paths:
                # No duplicate folder paths
                continue

            self.scan_shared_folder(folder_path, old_mtimes, old_files)

            self.processed_share_names.add(virtual_name)
            self.processed_share_paths.add(folder_path)

        # Save data to databases
        Shares.close_shares(self.share_dbs)
        self.set_shares(permission_level, files=self.files, streams=self.streams, mtimes=self.mtimes)

        for dictionary in (self.files, self.streams, self.mtimes):
            dictionary.clear()

        gc.collect()

    @classmethod
    def is_hidden(cls, folder, filename=None, entry=None):
        """Stop sharing any dot/hidden folders/files."""

        if filename is None:
            last_folder = os.path.basename(folder)

            if last_folder.startswith(".") or last_folder in cls.HIDDEN_FOLDER_NAMES:
                return True

        # If we're asked to check a file, we exclude it if it starts with a dot
        if filename is not None and filename.startswith("."):
            return True

        # Check if file is marked as hidden on Windows
        if sys.platform == "win32":
            if len(folder) == 3 and folder[1] == ":" and folder[2] == os.sep:
                # Root folders are marked as hidden, but we allow scanning them
                return False

            if entry is None:
                if filename is not None:
                    entry_stat = os.stat(encode_path(os.path.join(folder, filename)))
                else:
                    entry_stat = os.stat(encode_path(folder))
            else:
                entry_stat = entry.stat()

            return entry_stat.st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN

        return False

    def scan_shared_folder(self, shared_folder_path, old_mtimes, old_files):
        """Scan a shared folder for all subfolders, files and their metadata."""

        folder_paths = deque([shared_folder_path])

        while folder_paths:
            folder_path = folder_paths.pop()
            virtual_folder_path = self.real2virtual(folder_path)

            if virtual_folder_path in self.streams:
                # Sharing a folder twice, no go
                continue

            file_list = []

            try:
                with os.scandir(encode_path(folder_path, prefix=False)) as entries:
                    for entry in entries:
                        basename = entry.name.decode("utf-8", "replace")
                        path = os.path.join(folder_path, basename)

                        if entry.is_dir():
                            if self.is_hidden(path, entry=entry):
                                continue

                            self.current_folder_count += 1
                            folder_paths.append(path)

                            if not self.current_folder_count % 100:
                                self.queue.put(self.current_folder_count)
                            continue

                        try:
                            if self.is_hidden(folder_path, basename, entry):
                                continue

                            file_stat = entry.stat()
                            file_index = self.current_file_index
                            self.mtimes[path] = file_mtime = file_stat.st_mtime
                            virtual_file_path = f"{virtual_folder_path}\\{basename}"

                            if not self.rebuild and file_mtime == old_mtimes.get(path) and path in old_files:
                                full_path_file_data = old_files[path]
                                full_path_file_data[0] = virtual_file_path  # Virtual name might have changed
                            else:
                                full_path_file_data = self.get_file_info(virtual_file_path, path, file_stat)

                            basename_file_data = full_path_file_data[:]
                            basename_file_data[0] = basename
                            file_list.append(basename_file_data)

                            for k in set(virtual_file_path.lower().translate(TRANSLATE_PUNCTUATION).split()):
                                self.word_index[k].append(file_index)

                            self.files[path] = full_path_file_data
                            self.current_file_index += 1

                        except OSError as error:
                            self.queue.put(
                                ScannerLogMessage(
                                    _("Error while scanning file %(path)s: %(error)s"),
                                    {"path": path, "error": error}
                                )
                            )

            except OSError as error:
                self.queue.put(
                    ScannerLogMessage(
                        _("Error while scanning folder %(path)s: %(error)s"),
                        {"path": folder_path, "error": error}
                    )
                )

            self.streams[virtual_folder_path] = self.get_folder_stream(file_list)

    def get_audio_tag(self, encoded_file_path, size):

        parser_class = TinyTag._get_parser_for_filename(encoded_file_path)  # pylint: disable=protected-access

        if parser_class is None:
            return None

        with open(encoded_file_path, "rb") as file_handle:
            tag = parser_class(file_handle, size)
            tag.load(tags=False, duration=True, image=False)

        return tag

    def get_file_info(self, virtual_file_path, file_path, file_stat):
        """Get file metadata."""

        tag = None
        quality = None
        duration = None
        encoded_file_path = encode_path(file_path)
        size = file_stat.st_size

        # We skip metadata scanning of files without meaningful content
        if size > 128:
            try:
                tag = self.get_audio_tag(encoded_file_path, size)

            except Exception as error:
                self.queue.put(
                    ScannerLogMessage(
                        _("Error while scanning metadata for file %(path)s: %(error)s"),
                        {"path": file_path, "error": error}
                    )
                )

        if tag is not None:
            bitrate = tag.bitrate
            samplerate = tag.samplerate
            bitdepth = tag.bitdepth
            duration = tag.duration

            if bitrate is not None:
                bitrate = int(bitrate + 0.5)  # Round the value with minimal performance loss

                if not UINT32_LIMIT > bitrate > 0:
                    bitrate = None

            if samplerate is not None:
                samplerate = int(samplerate)

                if not UINT32_LIMIT > samplerate > 0:
                    samplerate = None

            if bitdepth is not None:
                bitdepth = int(bitdepth)

                if not UINT32_LIMIT > bitdepth > 0:
                    bitdepth = None

            if duration is not None:
                duration = int(duration)

                if not UINT32_LIMIT > duration >= 0:
                    duration = None

            quality = (bitrate, int(tag.is_vbr), samplerate, bitdepth)

        return [virtual_file_path, size, quality, duration]

    @staticmethod
    def get_folder_stream(file_list):
        """Pack all files and metadata in folder."""

        stream = bytearray()
        stream += FileListMessage.pack_uint32(len(file_list))

        for fileinfo in file_list:
            stream += FileListMessage.pack_file_info(fileinfo)

        return bytes(stream)


class Shares:
    __slots__ = ("share_dbs", "requested_share_times", "initialized", "rescanning", "compressed_shares",
                 "share_db_paths", "file_path_index", "_scanner_process")

    def __init__(self):

        self.share_dbs = {}
        self.requested_share_times = {}
        self.initialized = False
        self.rescanning = False
        self.compressed_shares = {
            PermissionLevel.PUBLIC: SharedFileListResponse(permission_level=PermissionLevel.PUBLIC),
            PermissionLevel.BUDDY: SharedFileListResponse(permission_level=PermissionLevel.BUDDY),
            PermissionLevel.TRUSTED: SharedFileListResponse(permission_level=PermissionLevel.TRUSTED),
            PermissionLevel.BANNED: SharedFileListResponse(permission_level=PermissionLevel.BANNED)
        }
        self.share_db_paths = {
            "words": os.path.join(config.data_folder_path, "words.dbn"),
            "public_files": os.path.join(config.data_folder_path, "publicfiles.dbn"),
            "public_mtimes": os.path.join(config.data_folder_path, "publicmtimes.dbn"),
            "public_streams": os.path.join(config.data_folder_path, "publicstreams.dbn"),
            "buddy_files": os.path.join(config.data_folder_path, "buddyfiles.dbn"),
            "buddy_mtimes": os.path.join(config.data_folder_path, "buddymtimes.dbn"),
            "buddy_streams": os.path.join(config.data_folder_path, "buddystreams.dbn"),
            "trusted_files": os.path.join(config.data_folder_path, "trustedfiles.dbn"),
            "trusted_mtimes": os.path.join(config.data_folder_path, "trustedmtimes.dbn"),
            "trusted_streams": os.path.join(config.data_folder_path, "trustedstreams.dbn")
        }
        self.file_path_index = ()

        self._scanner_process = None

        for event_name, callback in (
            ("folder-contents-request", self._folder_contents_request),
            ("quit", self._quit),
            ("server-disconnect", self._server_disconnect),
            ("server-login", self._server_login),
            ("shared-file-list-request", self._shared_file_list_request),
            ("shares-ready", self._shares_ready),
            ("start", self._start)
        ):
            events.connect(event_name, callback)

    def _start(self):

        rescan_startup = (config.sections["transfers"]["rescanonstartup"]
                          and not config.need_config())

        self.convert_shares()
        self.rescan_shares(init=True, rescan=rescan_startup)

    def _quit(self):

        self.close_shares(self.share_dbs)
        self.initialized = False

        if self._scanner_process is not None:
            self._scanner_process.terminate()

    def _server_login(self, msg):
        if msg.success:
            self.send_num_shared_folders_files()

    def _server_disconnect(self, _msg):
        self.requested_share_times.clear()

    # Shares-related Actions #

    @classmethod
    def create_db_file(cls, db_path):
        cls.remove_db_file(db_path)
        return Database(encode_path(db_path))

    @staticmethod
    def remove_db_file(db_path):

        db_path_encoded = encode_path(db_path)

        if os.path.isfile(db_path_encoded):
            os.remove(db_path_encoded)

        elif os.path.isdir(db_path_encoded):
            import shutil
            shutil.rmtree(db_path_encoded)

    def virtual2real(self, virtual_path):

        for shares in (
            config.sections["transfers"]["shared"],
            config.sections["transfers"]["buddyshared"],
            config.sections["transfers"]["trustedshared"]
        ):
            for virtual_name, folder_path, *_unused in shares:
                if virtual_path == virtual_name:
                    return folder_path

                if virtual_path.startswith(virtual_name + "\\"):
                    real_path = folder_path.rstrip(os.sep) + virtual_path[len(virtual_name):].replace("\\", os.sep)
                    return real_path

        return "__INVALID_SHARE__" + virtual_path

    def convert_shares(self):

        def _convert_share(shared_folder):
            if isinstance(shared_folder, tuple):
                virtual, shared_folder, *_unused = shared_folder
                virtual = str(virtual)
                shared_folder = str(shared_folder)
            else:
                # Convert fs-based shared to virtual shared (pre-1.4.0)
                virtual = str(shared_folder).replace("/", "_").replace("\\", "_").strip("_")
                log.add("Renaming shared folder '%s' to '%s'. A rescan of your share is required.",
                        (shared_folder, virtual))

            # Remove slashes from share name to avoid path conflicts
            virtual = virtual.replace("/", "_").replace("\\", "_")

            if shared_folder:
                shared_folder = os.path.normpath(shared_folder)

            return virtual, shared_folder

        config.sections["transfers"]["shared"] = [_convert_share(x)
                                                  for x in config.sections["transfers"]["shared"]]
        config.sections["transfers"]["buddyshared"] = [_convert_share(x)
                                                       for x in config.sections["transfers"]["buddyshared"]]

        # Remove old share databases (pre-3.3.0)
        for destination in (
            "wordindex", "fileindex", "files", "mtimes", "streams",
            "buddywordindex", "buddyfileindex", "buddyfiles", "buddymtimes", "buddystreams"
        ):
            file_path = os.path.join(config.data_folder_path, f"{destination}.db")

            try:
                self.remove_db_file(file_path)

            except OSError as error:
                log.add_debug("Failed to remove old share database %s: %s", (file_path, error))

    @classmethod
    def load_shares(cls, share_dbs, share_db_paths, destinations=None):

        exception = None

        for destination, db_path in share_db_paths.items():
            if destinations and destination not in destinations:
                continue

            try:
                share_dbs[destination] = Database(encode_path(db_path), overwrite=False)

            except Exception as error:
                exception = error
                cls.remove_db_file(db_path)

        if exception:
            cls.close_shares(share_dbs)
            raise exception

    def file_is_shared(self, username, virtual_path, real_path):

        log.add_transfer("Checking if file is shared: %s with real path %s",
                         (virtual_path, real_path))

        public_shared_files = self.share_dbs.get("public_files")
        buddy_shared_files = self.share_dbs.get("buddy_files")
        trusted_shared_files = self.share_dbs.get("trusted_files")
        file_is_shared = False
        size = None

        if not real_path.startswith("__INVALID_SHARE__"):
            if public_shared_files is not None and real_path in public_shared_files:
                file_is_shared = True
                _file_name, size, *_unused = public_shared_files[real_path]

            elif (buddy_shared_files is not None and username in core.buddies.users
                    and real_path in buddy_shared_files):
                file_is_shared = True
                _file_name, size, *_unused = buddy_shared_files[real_path]

            elif trusted_shared_files is not None:
                user_data = core.buddies.users.get(username)

                if user_data and user_data.is_trusted and real_path in trusted_shared_files:
                    file_is_shared = True
                    _file_name, size, *_unused = trusted_shared_files[real_path]

        if not file_is_shared:
            log.add_transfer("File is not present in the database of shared files, not sharing: "
                             "%s with real path %s", (virtual_path, real_path))
            return False, size

        return True, size

    def check_user_permission(self, username, ip_address=None):
        """Check if this user is banned, geoip-blocked, and which shares it is
        allowed to access based on transfer and shares settings."""

        if core.network_filter.is_user_banned(username) or core.network_filter.is_user_ip_banned(username, ip_address):
            if config.sections["transfers"]["usecustomban"]:
                ban_message = config.sections["transfers"]["customban"]
                return PermissionLevel.BANNED, ban_message

            return PermissionLevel.BANNED, ""

        user_data = core.buddies.users.get(username)

        if user_data:
            if user_data.is_trusted:
                return PermissionLevel.TRUSTED, ""

            return PermissionLevel.BUDDY, ""

        if ip_address is None or not config.sections["transfers"]["geoblock"]:
            return PermissionLevel.PUBLIC, ""

        country_code = core.network_filter.get_country_code(ip_address)

        # Please note that all country codes are stored in the same string at the first index
        # of an array, separated by commas (no idea why this decision was made...)

        if country_code and config.sections["transfers"]["geoblockcc"][0].find(country_code) >= 0:
            if config.sections["transfers"]["usecustomgeoblock"]:
                ban_message = config.sections["transfers"]["customgeoblock"]
                return PermissionLevel.BANNED, ban_message

            return PermissionLevel.BANNED, ""

        return PermissionLevel.PUBLIC, ""

    def get_shared_folders(self):

        shared_public_folders = config.sections["transfers"]["shared"]
        shared_buddy_folders = config.sections["transfers"]["buddyshared"]
        shared_trusted_folders = config.sections["transfers"]["trustedshared"]

        return shared_public_folders, shared_buddy_folders, shared_trusted_folders

    def get_normalized_virtual_name(self, virtual_name, shared_folders=None):

        if shared_folders is None:
            public_shares, buddy_shares, trusted_shares = self.get_shared_folders()
            shared_folders = (public_shares + buddy_shares + trusted_shares)

        # Provide a default name for root folders
        if not virtual_name:
            virtual_name = "Shared"

        # Remove slashes from share name to avoid path conflicts
        virtual_name = virtual_name.replace("/", "_").replace("\\", "_").strip(' "')
        new_virtual_name = str(virtual_name)

        # Check if virtual share name is already in use
        counter = 1
        while new_virtual_name in (x[0] for x in shared_folders):
            new_virtual_name = f"{virtual_name}{counter}"
            counter += 1

        return new_virtual_name

    def add_share(self, folder_path, permission_level=PermissionLevel.PUBLIC, virtual_name=None,
                  share_groups=None, validate_path=True):

        if validate_path and not os.access(encode_path(folder_path), os.R_OK):
            return None

        if share_groups is None:
            share_groups = self.get_shared_folders()

        # Remove previous share with same path if present
        core.shares.remove_share(folder_path, share_groups=share_groups)

        public_shares, buddy_shares, trusted_shares = share_groups
        virtual_name = core.shares.get_normalized_virtual_name(
            virtual_name or os.path.basename(folder_path),
            shared_folders=(public_shares + buddy_shares + trusted_shares)
        )
        permission_level_shares = {
            PermissionLevel.PUBLIC: public_shares,
            PermissionLevel.BUDDY: buddy_shares,
            PermissionLevel.TRUSTED: trusted_shares
        }
        shares = permission_level_shares.get(permission_level)

        shares.append((virtual_name, os.path.normpath(folder_path)))
        return virtual_name

    def remove_share(self, virtual_name_or_folder_path, share_groups=None):

        if share_groups is None:
            share_groups = self.get_shared_folders()

        normalized_folder_path = os.path.normpath(virtual_name_or_folder_path)

        for shares in share_groups:
            for virtual_name, folder_path in shares:
                if (virtual_name_or_folder_path in (virtual_name, folder_path)
                        or folder_path == normalized_folder_path):
                    shares.remove((virtual_name, folder_path))
                    return True

        return False

    @staticmethod
    def close_shares(share_dbs):

        for destination in share_dbs.copy():
            database = share_dbs.pop(destination, None)

            if database is not None:
                database.close()

    def send_num_shared_folders_files(self):
        """Send number of publicly shared files to the server."""

        if self.rescanning:
            return

        local_username = core.users.login_username
        num_shared_folders = len(self.share_dbs.get("public_streams", {}))
        num_shared_files = len(self.share_dbs.get("public_files", {}))

        if config.sections["transfers"]["reveal_buddy_shares"]:
            num_shared_folders += len(self.share_dbs.get("buddy_streams", {}))
            num_shared_files += len(self.share_dbs.get("buddy_files", {}))

        if config.sections["transfers"]["reveal_trusted_shares"]:
            num_shared_folders += len(self.share_dbs.get("trusted_streams", {}))
            num_shared_files += len(self.share_dbs.get("trusted_files", {}))

        core.send_message_to_server(SharedFoldersFiles(num_shared_folders, num_shared_files))

        if not local_username:
            # The shares module is initialized before the users module (to send updated share
            # stats to the server before watching our own username), and receives the login
            # event first, so local_username is still None when connecting. For this reason,
            # this check only happens after sending the SharedFoldersFiles server message.
            return

        # We've connected and rescanned our shares again. Fake a user stats message, since
        # server doesn't send updates for our own username after the first WatchUser message
        # response
        events.emit(
            "user-stats",
            GetUserStats(
                user=local_username,
                avgspeed=core.uploads.upload_speed, files=num_shared_files, dirs=num_shared_folders
            )
        )

    # Scanning #

    def rebuild_shares(self, use_thread=True):
        return self.rescan_shares(rebuild=True, use_thread=use_thread)

    def rescan_shares(self, init=False, rescan=True, rebuild=False, use_thread=True, force=False):

        if self.rescanning:
            return None

        if rescan and not force:
            # Verify all shares are mounted before allowing destructive rescan
            unavailable_shares = self.check_shares_available()

            if unavailable_shares:
                log.add(_("Rescan aborted due to unavailable shares: %s"), unavailable_shares)
                rescan = False

                events.emit("shares-unavailable", unavailable_shares)

                if not init:
                    return None

        # Hand over database control to the scanner process
        self.rescanning = True
        self.close_shares(self.share_dbs)
        self.file_path_index = ()

        events.emit("shares-preparing")

        share_groups = self.get_shared_folders()
        self._scanner_process, scanner_queue = self._build_scanner_process(share_groups, init, rescan, rebuild)
        self._scanner_process.start()

        if use_thread:
            Thread(
                target=self._process_scanner, args=(scanner_queue, events.emit_main_thread),
                name="ProcessShareScanner", daemon=True
            ).start()
            return None

        return self._process_scanner(scanner_queue)

    def check_shares_available(self):

        share_groups = self.get_shared_folders()
        unavailable_shares = []

        for share in share_groups:
            for virtual_name, folder_path, *_unused in share:
                if not os.access(encode_path(folder_path), os.R_OK):
                    unavailable_shares.append((virtual_name, folder_path))

        return unavailable_shares

    def _build_scanner_process(self, share_groups=None, init=False, rescan=True, rebuild=False):

        import multiprocessing

        context = multiprocessing.get_context(method="spawn")
        scanner_queue = context.Queue()
        scanner_obj = Scanner(
            scanner_queue,
            share_groups,
            self.share_db_paths,
            init,
            rescan,
            rebuild,
            reveal_buddy_shares=config.sections["transfers"]["reveal_buddy_shares"],
            reveal_trusted_shares=config.sections["transfers"]["reveal_trusted_shares"]
        )
        scanner = context.Process(target=scanner_obj.run, daemon=True)
        return scanner, scanner_queue

    def _process_scanner(self, scanner_queue, emit_event=None):

        successful = True

        while self._scanner_process.is_alive():
            # Cooldown
            time.sleep(0.05)

            while not scanner_queue.empty():
                item = scanner_queue.get()

                if item == ScannerState.FAILURE:
                    successful = False
                    break

                if isinstance(item, int):
                    if emit_event is not None:
                        emit_event("shares-scanning", item)

                elif isinstance(item, ScannerLogMessage):
                    log.add(item.msg, item.msg_args)

                elif isinstance(item, tuple):
                    self.file_path_index = item

                elif isinstance(item, SharedFileListResponse):
                    self.compressed_shares[item.permission_level] = item

                elif item == ScannerState.RESCANNING:
                    if emit_event is not None:
                        emit_event("shares-scanning")

                elif item == ScannerState.INITIALIZED:
                    self.initialized = True

        self._scanner_process = None

        if emit_event is not None:
            emit_event("shares-ready", successful)

        return successful

    def _shares_ready(self, successful):

        # Scanning done, load shares in the main process again
        if successful:
            try:
                self.load_shares(
                    self.share_dbs, self.share_db_paths, destinations={
                        "words", "public_files", "public_streams", "buddy_files", "buddy_streams",
                        "trusted_files", "trusted_streams"
                    })

            except Exception:
                successful = False

        self.rescanning = False

        if not successful:
            self.file_path_index = ()
            return

        self.send_num_shared_folders_files()

    # Network Messages #

    def _shared_file_list_request(self, msg):
        """Peer code 4."""

        username = msg.username
        request_time = time.monotonic()

        if username in self.requested_share_times and request_time < self.requested_share_times[username] + 0.4:
            # Ignoring request, because it's less than half a second since the
            # last one by this user
            return

        self.requested_share_times[username] = request_time

        log.add(_("User %(user)s is browsing your list of shared files"), {"user": username})

        ip_address, _port = msg.addr
        permission_level, _reject_reason = self.check_user_permission(username, ip_address)
        shares_list = self.compressed_shares.get(permission_level)
        shares_list.username = shares_list.sock = None  # Reset previous connection

        core.send_message_to_peer(username, shares_list)

    def _folder_contents_request(self, msg):
        """Peer code 36."""

        ip_address, _port = msg.addr
        username = msg.username
        folder_path = msg.dir
        permission_level, _reject_reason = self.check_user_permission(username, ip_address)
        folder_data = None

        if permission_level != PermissionLevel.BANNED:
            folder_data = self.share_dbs.get("public_streams", {}).get(folder_path)

            if (folder_data is None
                    and (config.sections["transfers"]["reveal_buddy_shares"]
                         or permission_level in {PermissionLevel.BUDDY, PermissionLevel.TRUSTED})):
                folder_data = self.share_dbs.get("buddy_streams", {}).get(folder_path)

            if (folder_data is None
                    and (config.sections["transfers"]["reveal_trusted_shares"]
                         or permission_level == PermissionLevel.TRUSTED)):
                folder_data = self.share_dbs.get("trusted_streams", {}).get(folder_path)

        core.send_message_to_peer(
            username, FolderContentsResponse(directory=folder_path, token=msg.token, shares=folder_data))
